#import <Foundation/Foundation.h>
#import "GHDelegate.h"

@protocol SBJsonStreamParserAdapterDelegate;
@class SBJsonStreamParser;
@class SBJsonStreamParserAdapter;

@class GHRepository;

/**
 * The superclass of Github API resources. Handles common connection-related tasks (and a delegate).
 *
 * Subclasses are meant to override:
 *
 * * apiPathWithParameters: - to provide the Github API path to the resource
 * * parser:foundObject: and/or parser:foundArray: - to collect the incoming data from the Github API
 */
@interface GHAbstractResource : NSObject <SBJsonStreamParserAdapterDelegate> {
  GHRepository *_repository;
  NSURLConnection *_githubConnection;
  SBJsonStreamParser *_jsonParser;
  SBJsonStreamParserAdapter *_jsonParserAdapter;
}

@property (nonatomic, readonly, strong) GHRepository *repository;
@property (nonatomic, weak) id<GHDelegate> delegate;

# pragma mark - Class methods

/**
 * \return the Github API base URL
 */
+ (NSString *)baseURL;

/**
 * \return the timestamp format used by Github
 */
+ (NSString *)timestampFormat;

/**
 * Creates an NSURLRequest for the Github API at the given path.
 *
 * \param path the path of the Github API resource. Must begin with a '/'.
 * \return a request to the Github API at the given path.
 */
+ (NSURLRequest *)requestForPath:(NSString *)path;

/**
 * Creates an NSURLRequest for the Github API at the given path with an Authorization header created for the given user.
 *
 * \param path the path of the Github API resource. Must begin with a '/'.
 * \param user the user authorizing this request
 * \return an authorized request to the Github API at the given path.
 */
//+ (NSURLRequest *)requestForPath:(NSString *)path onBehalfOfUser:(GHUser *)user;

# pragma mark - Instance methods

/**
 * Cancels the given connection.
 *
 * Should be extended by a subclass to perform any other cleanup or error messaging related to the cancellation. Make
 * sure to call super.
 *
 * \param connection the NSURLConnection that should be canceled.
 * \param response the HTTP response given that caused this cancellation.
 */
- (void)cancelConnection:(NSURLConnection *)connection withHTTPResponse:(NSHTTPURLResponse *)response;

/**
 * Cancels the given connection.
 *
 * Should be extended by a subclass to perform any other cleanup or error messaging related to the cancellation. Make
 * sure to call super.
 *
 * \param connection the NSURLConnection that should be canceled.
 * \param response the non-HTTP response given that caused this cancellation.
 */
- (void)cancelConnection:(NSURLConnection *)connection withResponse:(NSURLResponse *)response;

# pragma mark - Github interaction

/**
 * Returns the Github API path to this resource.
 *
 * Should be overridden by a subclass to provide the actual path.
 *
 * \return the Github API path to this resource, with the given query parameters.
 */
- (NSString *)apiPath;

/**
 * Starts a fetch of the data at the given path from the Github API. The delegate will be notified when the fetch is
 * complete.
 *
 * If the data at the given path is private, the Github credentials provided in the user property will be used to
 * authenticate the request.
 *
 * \param path the Github API path to fetch data from
 * \param parameters any query parameters to include with the request. May be nil.
 */
- (void)fetchDataAtApiPath:(NSString *)path withParameters:(NSDictionary *)parameters;

/**
 * Starts a fetch of this resource's data from the Github API. The delegate will be notified when the fetch is complete.
 *
 * If the repository on this object is private, the Github credentials provided in the user property will be used to
 * authenticate the request.
 *
 * Should be called from a subclass's explicitly named fetch method if it has one, e.g. GHTree#fetchLeaves =>
 * [self fetchDataWithParameters:{recursive=>1}];
 *
 * \param parameters any query parameters to include with the request. May be nil.
 */
- (void)fetchDataWithParameters:(NSDictionary *)parameters;

/**
 * Starts a fetch of this resource's data from the Github API. The delegate will be notified when the fetch is complete.
 *
 * If the repository on this object is private, the Github credentials provided in the user property will be used to
 * authenticate the request.
 *
 * Should be called from a subclass's explicitly named fetch method if it has one, e.g. GHBlob#fetchContent =>
 * [self fetchData];
 */
- (void)fetchData;

/**
 * Returns the Github API URL to this resource.
 *
 * \return the Github API URL to this resource
 */
- (NSString *)apiURL;

@end
