#import "SBJson.h"
#import "GHAbstractResource.h"
#import "GHRepository.h"
#import "NSString+Base64.h"
#import "GHErrors.h"

@implementation GHAbstractResource

@synthesize repository = _repository;
@synthesize delegate = _delegate;

# pragma mark - Class methods

+ (NSString *)baseURL {
  return @"https://api.github.com";
}

+ (NSString *)timestampFormat {
  return @"y'-'MM'-'dd'T'HH':'mm':'ssZ";
}

+ (NSURLRequest *)requestForPath:(NSString *)path {
  NSString *fullPath = [self baseURL];
  if (path) {
    fullPath = [fullPath stringByAppendingString:path];
  }
  return [NSURLRequest requestWithURL:[NSURL URLWithString:fullPath]];
}

/*+ (NSURLRequest *)requestForPath:(NSString *)path withCredentials:(GHCredentials *)credentials {
  NSURLRequest *request = [self requestForPath:path];
  if (user) {
    // create a base64-encoded string in the format username:password
    NSString *encodedLoginData = [[NSString stringWithFormat:@"%@:%@", credentials.login, credentials.password] encodeBase64];
    // add the encoded credentials to Authorization
    NSMutableURLRequest *mutableRequest = [request mutableCopy];
    [mutableRequest addValue:[NSString stringWithFormat:@"Basic %@", encodedLoginData]
          forHTTPHeaderField:@"Authorization"];
    request = mutableRequest;
  }
  return request;
}*/

- (id)init
{
  self = [super init];
  if (self) {
    _jsonParserAdapter = [[SBJsonStreamParserAdapter alloc] init];
    _jsonParserAdapter.delegate = self;
  }
  return self;
}

# pragma mark - Instance methods

- (void)cancelConnection:(NSURLConnection *)connection withHTTPResponse:(NSHTTPURLResponse *)response {
  NSLog(@"Cancelling connection.");
  [connection cancel];
  // Inform the delegate of what happened
  if (self.delegate) {
    NSDictionary *headers = [response allHeaderFields];
    int status = [[headers objectForKey:@"Status"] intValue];
    NSLog(@"Connection cancelled due to status %d", status);
    if ([self.delegate respondsToSelector:@selector(githubObject:didCancelFetchWithError:)]) {
      NSUInteger errorCode = 0;
      NSString *description = nil;
      NSString *recoverySuggestion = nil;
      NSArray *recoveryOptions = nil;
      if (status == 304) {
        errorCode = GHConnectionNotModifiedError;
        description = NSLocalizedStringFromTable(@"github.connection.not_modified.description", GITHUB_STRINGS_TABLE, @"GHConnectionNotModifiedError NSLocalizedDescriptionKey");
        recoverySuggestion = NSLocalizedStringFromTable(@"github.connection.not_modified.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionNotModifiedError NSLocalizedRecoverySuggestionErrorKey");
        recoveryOptions = [NSArray arrayWithObject:NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue")];
      } else if (status == 401) {
        errorCode = GHConnectionUnauthorizedError;
        description = NSLocalizedStringFromTable(@"github.connection.unauthorized.description", GITHUB_STRINGS_TABLE, @"GHConnectionUnauthorizedError NSLocalizedDescriptionKey");
        recoverySuggestion = NSLocalizedStringFromTable(@"github.connection.unauthorized.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionUnauthorizedError NSLocalizedRecoverySuggestionErrorKey");
        recoveryOptions = [NSArray arrayWithObject:NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue")];
      } else if (status == 403) {
        errorCode = GHConnectionForbiddenError;
        description = NSLocalizedStringFromTable(@"github.connection.forbidden.description", GITHUB_STRINGS_TABLE, @"GHConnectionForbiddenError NSLocalizedDescriptionKey");
        recoverySuggestion = NSLocalizedStringFromTable(@"github.connection.forbidden.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionForbiddenError NSLocalizedRecoverySuggestionErrorKey");
        recoveryOptions = [NSArray arrayWithObject:NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue")];
      } else if (status == 404) {
        errorCode = GHConnectionNotFoundError;
        description = NSLocalizedStringFromTable(@"github.connection.not_found.description", GITHUB_STRINGS_TABLE, @"GHConnectionNotFoundError NSLocalizedDescriptionKey");
        recoverySuggestion = NSLocalizedStringFromTable(@"github.connection.not_found.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionNotFoundError NSLocalizedRecoverySuggestionErrorKey");
        recoveryOptions = [NSArray arrayWithObject:NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue")];
      } else {
        errorCode = GHConnectionUnknownError;
        description = NSLocalizedStringFromTable(@"github.connection.unknown.description", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedDescriptionKey");
        recoverySuggestion = NSLocalizedStringFromTable(@"github.connection.unknown.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoverySuggestionErrorKey");
        recoveryOptions = [NSArray arrayWithObjects:NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.try_again", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Try the request again"),
                           NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.go_back", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Go back and try something different"), nil];
      }
      NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:description, NSLocalizedDescriptionKey,
                                recoverySuggestion, NSLocalizedRecoverySuggestionErrorKey,
                                recoveryOptions, NSLocalizedRecoveryOptionsErrorKey, nil];
      NSError *error = [NSError errorWithDomain:GITHUB_ERROR_DOMAIN code:errorCode userInfo:userInfo];
      [self.delegate githubObject:self didCancelFetchWithError:error];
    }
  }
}

- (void)cancelConnection:(NSURLConnection *)connection withResponse:(NSURLResponse *)response {
  [connection cancel];
  // Inform the delegate of what happened
  if ([self.delegate respondsToSelector:@selector(githubObject:didCancelFetchWithError:)]) {
    NSUInteger errorCode = GHConnectionUnknownError;
    NSString *description = NSLocalizedStringFromTable(@"github.connection.unknown.description", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedDescriptionKey");
    NSString *recoverySuggestion = NSLocalizedStringFromTable(@"github.connection.unknown.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoverySuggestionErrorKey");
    NSArray *recoveryOptions = [NSArray arrayWithObjects:NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.try_again", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Try the request again"),
                                NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.go_back", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Go back and try something different"), nil];
    NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:description, NSLocalizedDescriptionKey,
                              recoverySuggestion, NSLocalizedRecoverySuggestionErrorKey,
                              recoveryOptions, NSLocalizedRecoveryOptionsErrorKey, nil];
    NSError *error = [NSError errorWithDomain:GITHUB_ERROR_DOMAIN code:errorCode userInfo:userInfo];
    [self.delegate githubObject:self didCancelFetchWithError:error];
  }
}

# pragma mark - Github interaction

- (NSString *)apiPath {
  return @"/";
}

- (void)fetchDataAtApiPath:(NSString *)path withParameters:(NSDictionary *)parameters {
  if (path) {
    // TODO: authentication for private resource
    if (parameters) {
      NSMutableString *queryString = [NSMutableString string];
      __block BOOL first = YES;
      [parameters enumerateKeysAndObjectsUsingBlock:^(id key, id obj, BOOL *stop) {
        if (first) {
          [queryString appendString:@"?"];
          first = NO;
        } else {
          [queryString appendString:@"&"];
        }
        [queryString appendString:[NSString stringWithFormat:@"%@=%@", [key description], [obj description]]];
      }];
      path = [path stringByAppendingString:queryString];
    }
    NSLog(@"connection to the Github API at %@", path);
    NSURLRequest *request = [self.class requestForPath:path];
    // make the connection
    _githubConnection = [NSURLConnection connectionWithRequest:request delegate:self];
    if (!_githubConnection) {
      NSLog(@"Connection to %@ failed to create.", [[request URL] description]);
      // TODO: notify delegate
    }
  }
}

- (void)fetchDataWithParameters:(NSDictionary *)parameters {
  [self fetchDataAtApiPath:[self apiPath] withParameters:parameters];
}

- (void)fetchData {
  [self fetchDataWithParameters:nil];
}

- (NSString *)apiURL {
  return [NSString stringWithFormat:@"%@%@", [self.class baseURL], [self apiPath]];
}

# pragma mark - NSURLConnection delegate methods

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response {
  if (response) {
    // reset any existing connection
    _jsonParser = nil;
    
    if ([response isKindOfClass:[NSHTTPURLResponse class]]) {
      NSHTTPURLResponse *httpResponse = (NSHTTPURLResponse *)response;
      // TODO: Follow page links in Link header?
      NSDictionary *headers = [httpResponse allHeaderFields];
      NSString *status = [headers objectForKey:@"Status"];
      if ([status intValue] == 200) {
        // create new parser to avoid data conflicts
        _jsonParser = [[SBJsonStreamParser alloc] init];
        _jsonParser.delegate = _jsonParserAdapter;
      } else {
        NSLog(@"Non-successful response code %d, headers: %@", [status intValue], [headers description]);
        [self cancelConnection:connection withHTTPResponse:httpResponse];
      }
    } else {
      NSLog(@"non-http response received: %@", [response MIMEType]);
      [self cancelConnection:connection withResponse:response];
    }
  } else {
    NSLog(@"Ignoring nil response.");
  }
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data {
  if (data) {
    if (_jsonParser) {
      [_jsonParser parse:data];
    }
  } else {
    NSLog(@"expected to receive data, but did not receive any.");
  }
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error {
  _jsonParser = nil;
  _githubConnection = nil;
  NSLog(@"connection failed: %@", [error localizedDescription]);
  if ([self.delegate respondsToSelector:@selector(githubObject:connectionDidFailWithError:)]) {
    [self.delegate githubObject:self connectionDidFailWithError:error];
  }
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection {
  _jsonParser = nil;
  _githubConnection = nil;
  NSLog(@"connection finished.");
  if ([self.delegate respondsToSelector:@selector(githubObjectConnectionDidFinish:)]) {
    [self.delegate githubObjectConnectionDidFinish:self];
  }
}

# pragma mark - SBJsonStreamParserAdapter delegate methods

- (void)parser:(SBJsonStreamParser*)parser foundArray:(NSArray*)array {
}

- (void)parser:(SBJsonStreamParser*)parser foundObject:(NSDictionary*)dict {
}

@end
