#import "GHTreeLeaf.h"
#import "GHRepository.h"

/**
 * A Git blob as modeled in the Github API. A blob is a representation of a file at a particular point in time. A SHA
 * hash uniquely identifies a blob in its repository.
 *
 * The binary content of the blob is in #content. TODO: If the blob contains textual content, #encoding will hold the
 * text encoding necessary to decode the #content.
 *
 * This blob is read-only. TODO: To create or update a blob, call #mutableCopy to modify the blob.
 */
@interface GHBlob : GHTreeLeaf <NSCopying> {
  NSData *_content;
  NSStringEncoding _encoding;
}

# pragma mark - Attributes

/**
 * The binary content of the blob.
 */
@property (nonatomic, readonly, copy) NSData *content;

/**
 * The text encoding of the blob. If the blob is binary, this will be NSProprietaryEncoding.
 */
@property (nonatomic, readonly, assign) NSStringEncoding encoding;

/**
 * Encodes the blob content as a string using the blob's stored encoding.
 *
 * If this is a binary blob, this method should not be called. If it is, the default CString encoding will be used.
 *
 * \return the blob content encoded as a string using the blob's stored encoding
 */
- (NSString *)contentAsNSString;

/**
 * Indicates if this blob contains text or binary data.
 *
 * \return YES if this blob contains encoded text, NO if it contains another kind of data.
 */
- (BOOL)isText;

# pragma mark - Class constructors

/**
 * Creates a GHBlob with the given sha and starts a fetch of its content from Github.
 *
 * TODO: Authenticates the request with the given Github credentials. If the repository is open source, the user is not
 * necessary.
 *
 * The fetch is asynchronous and notifies the delegate when it is complete.
 *
 * \param repository the Github repository that owns this blob
 * \param sha the sha hash for this blob
 * \param delegate the delegate to notify when the content has been fetched
 * \return a new GHBlob that is in the process of fetching its content
 */
+ (id)fetchBlobContentWithRepository:(GHRepository *)repository sha:(NSString *)sha delegate:(id<GHDelegate>)delegate;

# pragma mark - Initializers

/**
 * Creates a blob with the bare minimum attributes to fetch its contents from Github.
 *
 * TODO: The request to fetch the blob content will be authenticated with the given credentials. If the repository is
 * open source, the credentials are not necessary.
 *
 * \param repository the Github repository that owns this blob
 * \param sha the sha hash for this blob
 * \return a new GHBlob without content
 */
- (id)initWithRepository:(GHRepository *)repository sha:(NSString *)sha;

/**
 * Creates a blob with the attributes found by the Github Tree API. The blob can fetch its contents from the same API.
 *
 * TODO: The request to fetch the blob content will be authenticated with the given credentials. If the repository is
 * open source, the user is not necessary.
 *
 * \param repository the Github repository that owns this blob
 * \param mode the file permissions for this blob
 * \param path the path information for this blob, including filename
 * \param sha the sha hash for this blob
 * \param size the file size of the blob
 * \return a new GHBlob without content
 */
- (id)initWithRepository:(GHRepository *)repository mode:(NSString *)mode path:(NSString *)path sha:(NSString *)sha size:(NSUInteger)size;

/**
 * Creates a complete blob with the given attributes.
 *
 * \param repository the Github repository that owns this blob
 * \param content the content of the blob
 * \param encoding the text encoding of the content (if it's text)
 * \param mode the file permissions for this blob
 * \param path the path information for this blob, including filename
 * \param sha the sha hash for this blob
 * \param size the file size of the blob
 * \return a new GHBlob without content
 */
- (id)initWithRepository:(GHRepository *)repository content:(NSData *)content encoding:(NSStringEncoding)encoding mode:(NSString *)mode path:(NSString *)path sha:(NSString *)sha size:(NSUInteger)size;

# pragma mark - Github interaction

/**
 * Fetches the contents of this blob, as represented by its SHA, from Github. Notifies the delegate when the fetch is
 * complete.
 */
- (void)fetchContent;

@end
