#import <Foundation/Foundation.h>

@class GHAbstractResource;

/**
 * Contains methods used to communicate Github connection-related events to a delegate.
 *
 * githubObjectConnectionDidFinish: notifies the delegate that the fetch has completed successfully and data is
 * available for consumption.
 *
 * githubObject:didCancelFetchWithError: is used when a non-successful response is received from Github, like 304 Not
 * Modified or 404 Not Found.
 *
 * githubObject:cannotFetchWithError: is used when the request could not be constructed, like if required repository
 * information was missing.
 *
 * githubObject:connectionDidFailWithError: is used for a connection error and implies Github was not reached.
 */
@protocol GHDelegate <NSObject>

/**
 * Called when the connection finished.
 *
 * \param githubObject the GHAbstractResource making the request
 */
- (void)githubObjectConnectionDidFinish:(GHAbstractResource *)githubObject;

@optional

/**
 * Called when the connection was canceled due to not getting back a success response from Github, like 304 Not Modified
 * or 404 Not Found.
 *
 * \param githubObject the GHAbstractResource making the request
 * \param error an error object explaining the issue and possible resolutions
 */
- (void)githubObject:(GHAbstractResource *)githubObject didCancelFetchWithError:(NSError *)error;

/**
 * Called when the connection was not started due to missing required information, like required repository information.
 *
 * \param githubObject the GHAbstractResource making the request
 * \param error an error object explaining the issue and possible resolutions
 */
- (void)githubObject:(GHAbstractResource *)githubObject cannotFetchWithError:(NSError *)error;

/**
 * Called when the connection failed due to an error in the connection. Implies Github was not reached.
 *
 * \param githubObject the GHAbstractResource making the request
 * \param error the error object explaining the failure
 */
- (void)githubObject:(GHAbstractResource *)githubObject connectionDidFailWithError:(NSError *)error;

@end
