#import "GHTreeLeaf.h"

/**
 * A Git tree as modeled in the Github API. A tree is a representation of the repository's file structure at a
 * particular point in time. A SHA hash uniquely identifies a tree in its repository.
 *
 * A tree in this framework composes a container for GHTreeLeafs. It implements NSFastEnumeration to provide a handy
 * interface for reaching those blobs.
 *
 * This tree is read-only. TODO: To create or update a tree, call #mutableCopy to modify the tree.
 */
@interface GHTree : GHTreeLeaf <NSCopying, NSFastEnumeration> {
  NSArray *_leaves;
}

# pragma mark - Attributes

/**
 * The GHTreeLeaf leaves of this tree.
 */
@property (nonatomic, readonly, copy) NSArray *leaves;

# pragma mark - Initializers

/**
 * Creates a tree with the bare minimum properties to fetch its leaves from Github.
 *
 * TODO: The request to fetch the tree leaves will be authenticated with the given credentials. If the repository is
 * open source, the credentials are not necessary.
 *
 * \param repository the Github repository that owns this tree
 * \param sha the sha hash for this tree
 * \return a new GHTree without leaves
 */
- (id)initWithRepository:(GHRepository *)repository sha:(NSString *)sha;

/**
 * Creates a subtree with the properties found by the Github Tree API. The tree can fetch its leaves from the same API.
 *
 * TODO: The request to fetch the tree leaves will be authenticated with the given credentials. If the repository is
 * open source, the credentials are not necessary.
 *
 * \param repository the Github repository that owns this tree
 * \param mode the file permissions for this tree
 * \param path the path to this tree
 * \param sha the sha hash for this tree
 * \return a new GHTree without leaves
 */
- (id)initWithRepository:(GHRepository *)repository mode:(NSString *)mode path:(NSString *)path sha:(NSString *)sha;

/**
 * Creates a complete tree with the given properties.
 *
 * \param repository the Github repository that owns this tree
 * \param leaves an array of GHTreeLeaf objects that this tree contains
 * \param mode the file permissions for this tree
 * \param path the path to this tree
 * \param sha the sha hash for this tree
 * \return a new GHTree with the given properties
 */
- (id)initWithRepository:(GHRepository *)repository leaves:(NSArray *)leaves mode:(NSString *)mode path:(NSString *)path sha:(NSString *)sha;

# pragma mark - Github interaction

/**
 * Fetches the immediate leaves of this tree, as represented by its SHA, from Github. Immediate leaves are those one
 * level underneath this tree, and could contain both blobs and subtrees. Notifies the delegate when the fetch is
 * complete.
 */
- (void)fetchImmediateLeaves;

/**
 * Fetches all the blob leaves of this tree, as represented by its SHA, from Github. Notifies the delegate when the
 * fetch is complete.
 */
- (void)fetchLeaves;

@end
