#import "GHAbstractResource.h"

/**
 * A Github user. Useful in several contexts:
 *
 * * obtaining user information - a username can provide a public profile for that user
 * * listing repositories - a username can list all public repositories for that user; username and password can list
 *                          all public and private repositories for that user.
 * * authenticating requests - for requests regarding private methods or resources, a username and password combination
 *                             are required.
 *
 * This user is read-only. To update a user, call #mutableCopy to modify the user.
 */
@interface GHUser : GHAbstractResource

# pragma mark - Attributes

@property (nonatomic, readonly, copy) NSString *avatarURL;
@property (nonatomic, readonly, copy) NSString *bio;
@property (nonatomic, readonly, copy) NSString *blog;
@property (nonatomic, readonly, assign) NSUInteger collaboratorCount;
@property (nonatomic, readonly, copy) NSString *company;
@property (nonatomic, readonly, strong) NSDate *createdAt;
@property (nonatomic, readonly, assign) NSUInteger diskUsage;
@property (nonatomic, readonly, copy) NSString *email;
@property (nonatomic, readonly, assign) NSUInteger followerCount;
@property (nonatomic, readonly, assign) NSUInteger followingCount;
/**
 * The URL to the user's HTML profile page.
 */
@property (nonatomic, readonly, copy) NSString *githubURL;
@property (nonatomic, readonly, copy) NSString *gravatarId;
@property (nonatomic, readonly, assign) BOOL hireable;
@property (nonatomic, readonly, copy) NSString *location;
@property (nonatomic, readonly, copy) NSString *login;
@property (nonatomic, readonly, copy) NSString *name;
@property (nonatomic, readonly, assign) NSUInteger ownedRepositoryCount;
/**
 * Information about the user's Github plan.
 *
 * Keys:
 * * name
 * * space
 * * collaborators (TODO: collaboratorCount)
 * * private_repos (TODO: privateRepositoryCount)
 */
@property (nonatomic, readonly, copy) NSDictionary *plan;
@property (nonatomic, readonly, assign) NSUInteger privateGistCount;
@property (nonatomic, readonly, assign) NSUInteger privateRepositoryCount;
@property (nonatomic, readonly, assign) NSUInteger publicGistCount;
@property (nonatomic, readonly, assign) NSUInteger publicRepositoryCount;

# pragma mark - Github API attribute names

@property (nonatomic, readonly, copy) NSString *avatar_url;
@property (nonatomic, readonly, assign) NSUInteger collaborators;
@property (nonatomic, readonly, strong) NSDate *created_at;
@property (nonatomic, readonly, assign) NSUInteger disk_usage;
@property (nonatomic, readonly, assign) NSUInteger followers;
@property (nonatomic, readonly, assign) NSUInteger following;
@property (nonatomic, readonly, copy) NSString *gravatar_id;
@property (nonatomic, readonly, copy) NSString *html_url;
@property (nonatomic, readonly, assign) NSUInteger owned_private_repos;
@property (nonatomic, readonly, assign) NSUInteger private_gists;
@property (nonatomic, readonly, assign) NSUInteger public_gists;
@property (nonatomic, readonly, assign) NSUInteger public_repos;
@property (nonatomic, readonly, assign) NSUInteger total_private_repos;

# pragma mark - Initializers

/**
 * Creates a user with a username to provide the location of a resource.
 *
 * \param login the username
 */
- (id)initWithLogin:(NSString *)login;

/**
 * Creates a user with the given attributes.
 *
 * \param attributes the desired attributes, where each key is a property name or the Github API property name
 */
- (id)initWithAttributes:(NSDictionary *)attributes;

# pragma mark - Github interaction

/**
 * Starts a fetch of a list of this user's repositories from the Github API. The delegate will be notified when the
 * fetch is complete, if the delegate is a GHUserDelegate.
 *
 * If the user has private repositories, the request must be authenticated (TODO).
 */
- (void)fetchRepositories;

@end
