#import "GHAbstractResourceTests.h"
#import <OCMock/OCMock.h>
#import "GHRepository.h"
#import "GHErrors.h"
#import "NSString+Base64.h"
#import "GHDelegate.h"

/**
 * Returns an NSError with the correct information to return the given localizedDescription for #localizedDescription.
 */
static id mockNSError(NSString *localizedDescription);

/**
 * Returns a mock NSURLResponse.
 */
static id mockNSURLResponse();

/**
 * Add delegate method interface to GHAbstractResource.
 */
@interface GHAbstractResource (GHAbstractResourceAsDelegate)
- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response;
- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data;
- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error;
- (void)connectionDidFinishLoading:(NSURLConnection *)connection;
- (void)parser:(SBJsonStreamParser*)parser foundArray:(NSArray*)array;
- (void)parser:(SBJsonStreamParser*)parser foundObject:(NSDictionary*)dict;
@end

/**
 * A subclass used to ensure the correct request is generated. Stores the request for later retrieval by the test.
 */
@interface GHAbstractResourceTester : GHAbstractResource
@property (nonatomic, strong) NSURLRequest *fetchRequest;
- (void)connection:(NSURLConnection *)connection didReceiveRequest:(NSURLRequest *)request;
@end
@implementation GHAbstractResourceTester
@synthesize fetchRequest = _fetchRequest;
- (void)connection:(NSURLConnection *)connection didReceiveRequest:(NSURLRequest *)request {
  self.fetchRequest = request;
}
@end

@implementation GHAbstractResourceTests

- (void)setUp
{
  [super setUp];
  abstractResource = [[GHAbstractResource alloc] init];
}

- (void)tearDown
{
  abstractResource = nil;
  [super tearDown];
}

# pragma mark - Attributes

/** repository tests */

- (void)testRepository_GHRepository {
  GHRepository *repo = [[GHRepository alloc] init];
  [abstractResource setValue:repo forKey:@"_repository"];
  STAssertEquals(repo, abstractResource.repository, @"should have stored repository");
}

- (void)testSetRepository_NoMethod {
  STAssertThrowsSpecificNamed([abstractResource performSelector:@selector(setRepository:)
                                                     withObject:[[GHRepository alloc] init]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setRepository: selector");
}

/** delegate tests */

- (void)testDelegate_GHDelegate {
  MockGHDelegate *delegate = [[MockGHDelegate alloc] init];
  [abstractResource setValue:delegate forKey:@"_delegate"];
  STAssertEquals(delegate, abstractResource.delegate, @"should have stored delegate");
}

- (void)testSetDelegate_GHDelegate_AssignedDelegate {
  MockGHDelegate *delegate = [[MockGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  STAssertEquals(delegate, [abstractResource valueForKey:@"_delegate"], @"should have stored delegate");
}

# pragma mark - Class methods

/** requestForPath: tests */

- (void)testRequestForPath_Nil_BaseURL {
  NSURLRequest *request = [GHAbstractResource requestForPath:nil];
  STAssertTrue([[[request URL] absoluteString] isEqualToString:@"https://api.github.com"],
               @"nil path should result in request to base URL");
}

- (void)testRequestForPath_Path_URLWithPath {
  NSURLRequest *request = [GHAbstractResource requestForPath:@"/"];
  STAssertTrue([[[request URL] absoluteString] isEqualToString:@"https://api.github.com/"],
               @"path should result in request to base URL with the given path");
}

/** requestForPath:onBehalfOfUser: tests /

- (void)testRequestForPathOnBehalfOfUser_NilPathNilUser_BaseURLNoAuthorization {
  NSURLRequest *request = [GHAbstractResource requestForPath:nil onBehalfOfUser:nil];
  STAssertTrue([[[request URL] absoluteString] isEqualToString:@"https://api.github.com"],
               @"nil path should result in request to base URL");
  STAssertNil([request valueForHTTPHeaderField:@"Authorization"], @"nil user should result in no Authorization header");
}

- (void)testRequestForPathOnBehalfOfUser_NilPathGithubUser_BaseURLWithAuthorization {
  GHUser *user = [[GHUser alloc] initWithUsername:@"tester" password:@"testing"];
  NSURLRequest *request = [GHAbstractResource requestForPath:nil onBehalfOfUser:user];
  STAssertTrue([[[request URL] absoluteString] isEqualToString:@"https://api.github.com"],
               @"nil path should result in request to base URL");
  NSString *encodedLoginString = [[NSString stringWithFormat:@"%@:%@", @"tester", @"testing"] encodeBase64];
  NSString *expected = [NSString stringWithFormat:@"Basic %@", encodedLoginString];
  STAssertTrue([[request valueForHTTPHeaderField:@"Authorization"] isEqualToString:expected],
               @"authorization header should have the correct data");
}

- (void)testRequestForPathOnBehalfOfUser_PathNilUser_URLWithPathNoAuthorization {
  NSURLRequest *request = [GHAbstractResource requestForPath:@"/" onBehalfOfUser:nil];
  STAssertTrue([[[request URL] absoluteString] isEqualToString:@"https://api.github.com/"],
               @"path should result in request to base URL with the given path");
  STAssertNil([request valueForHTTPHeaderField:@"Authorization"], @"nil user should result in no Authorization header");
}

- (void)testRequestForPathOnBehalfOfUser_PathGithubUser_URLWithPathAndAuthorization {
  GHUser *user = [[GHUser alloc] initWithUsername:@"tester" password:@"testing"];
  NSURLRequest *request = [GHAbstractResource requestForPath:@"/" onBehalfOfUser:user];
  STAssertTrue([[[request URL] absoluteString] isEqualToString:@"https://api.github.com/"],
               @"path should result in request to base URL with the given path");
  NSString *encodedLoginString = [[NSString stringWithFormat:@"%@:%@", @"tester", @"testing"] encodeBase64];
  NSString *expected = [NSString stringWithFormat:@"Basic %@", encodedLoginString];
  STAssertTrue([[request valueForHTTPHeaderField:@"Authorization"] isEqualToString:expected],
               @"authorization header should have the correct data");
}*/

# pragma mark - Instance methods

/** cancelConnection:withHTTPResponse: tests */

- (void)testCancelConnectionWithHTTPResponse_NilConnection_NoException {
  STAssertNoThrow([abstractResource cancelConnection:nil withHTTPResponse:nil],
                  @"should not have thrown for nil arguments");
}

- (void)testCancelConnectionWithHTTPResponse_NilDelegate_NoException {
  abstractResource.delegate = nil;
  STAssertNoThrow([abstractResource cancelConnection:mockNSURLConnection() withHTTPResponse:mockNSHTTPURLResponse(200)],
                  @"should not have thrown for nil delegate");
}

- (void)testCancelConnectionWithHTTPResponse_UnimplementedMethodDelegate_NoException {
  MockGHDelegate *delegate = [[MockGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  STAssertNoThrow([abstractResource cancelConnection:mockNSURLConnection() withHTTPResponse:mockNSHTTPURLResponse(200)],
                  @"should not have thrown for delegate not implementing githubObject:didCancelFetchWithError:");
}

- (void)testCancelConnectionWithHTTPResponse_Connection_CanceledConnection {
  id connection = [OCMockObject mockForClass:[NSURLConnection class]];
  [[connection expect] cancel];
  [abstractResource cancelConnection:connection withHTTPResponse:mockNSHTTPURLResponse(200)];
  STAssertNoThrow([connection verify], @"should have called cancel on the connection");
}

- (void)testCancelConnectionWithHTTPResponse_NotModifiedHTTPResponse_NotifiedDelegateResourceNotModified {
  MockCancelGHDelegate *delegate = [[MockCancelGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  [abstractResource cancelConnection:mockNSURLConnection() withHTTPResponse:mockNSHTTPURLResponse(304)];
  STAssertTrue(delegate.called, @"should have called githubObject:didCancelFetchWithError: on delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHConnectionNotModifiedError, [delegate.error code], @"should have indicated resource not modified as the error");
  NSDictionary *userInfo = [delegate.error userInfo];
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.not_modified.description", GITHUB_STRINGS_TABLE, @"GHConnectionNotModifiedError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.not_modified.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionNotModifiedError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(1U, [recoveryOptions count], @"should have included one recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct recovery option");
}

- (void)testCancelConnectionWithHTTPResponse_UnauthorizedHTTPResponse_NotifiedDelegateResourceUnauthorized {
  MockCancelGHDelegate *delegate = [[MockCancelGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  [abstractResource cancelConnection:mockNSURLConnection() withHTTPResponse:mockNSHTTPURLResponse(401)];
  STAssertTrue(delegate.called, @"should have called githubObject:didCancelFetchWithError: on delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHConnectionUnauthorizedError, [delegate.error code], @"should have indicated request unauthorized as the error");
  NSDictionary *userInfo = [delegate.error userInfo];
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unauthorized.description", GITHUB_STRINGS_TABLE, @"GHConnectionUnauthorizedError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unauthorized.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionUnauthorizedError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(1U, [recoveryOptions count], @"should have included one recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct recovery option");
}

- (void)testCancelConnectionWithHTTPResponse_ForbiddenHTTPResponse_NotifiedDelegateResourceForbidden {
  MockCancelGHDelegate *delegate = [[MockCancelGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  [abstractResource cancelConnection:mockNSURLConnection() withHTTPResponse:mockNSHTTPURLResponse(403)];
  STAssertTrue(delegate.called, @"should have called githubObject:didCancelFetchWithError: on delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHConnectionForbiddenError, [delegate.error code], @"should have indicated request forbidden as the error");
  NSDictionary *userInfo = [delegate.error userInfo];
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.forbidden.description", GITHUB_STRINGS_TABLE, @"GHConnectionForbiddenError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.forbidden.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionForbiddenError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(1U, [recoveryOptions count], @"should have included one recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct recovery option");
}

- (void)testCancelConnectionWithHTTPResponse_NotFoundHTTPResponse_NotifiedDelegateResourceNotFound {
  MockCancelGHDelegate *delegate = [[MockCancelGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  [abstractResource cancelConnection:mockNSURLConnection() withHTTPResponse:mockNSHTTPURLResponse(404)];
  STAssertTrue(delegate.called, @"should have called githubObject:didCancelFetchWithError: on delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHConnectionNotFoundError, [delegate.error code], @"should have indicated resource not found as the error");
  NSDictionary *userInfo = [delegate.error userInfo];
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.not_found.description", GITHUB_STRINGS_TABLE, @"GHConnectionNotFoundError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.not_found.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionNotFoundError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(1U, [recoveryOptions count], @"should have included one recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct recovery option");
}

- (void)testCancelConnectionWithHTTPResponse_NilHTTPResponse_NotifiedDelegateFetchCanceled {
  MockCancelGHDelegate *delegate = [[MockCancelGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  [abstractResource cancelConnection:mockNSURLConnection() withHTTPResponse:nil];
  STAssertTrue(delegate.called, @"should have called githubObject:didCancelFetchWithError: on delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHConnectionUnknownError, [delegate.error code], @"should have indicated an unknown error");
  NSDictionary *userInfo = [delegate.error userInfo];
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.description", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(2U, [recoveryOptions count], @"should have included two recovery options");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.try_again", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Try the request again") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct first recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.go_back", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Go back and try something different") isEqualToString:[recoveryOptions objectAtIndex:1]],
               @"should have included the correct second recovery option");
}

/** cancelConnection:withResponse: tests */

- (void)testCancelConnectionWithResponse_NilConnection_NoException {
  STAssertNoThrow([abstractResource cancelConnection:nil withResponse:nil],
                  @"should not have thrown for nil delegate");
}

- (void)testCancelConnectionWithResponse_NilDelegate_NoException {
  abstractResource.delegate = nil;
  STAssertNoThrow([abstractResource cancelConnection:mockNSURLConnection() withResponse:mockNSURLResponse()],
                  @"should not have thrown for nil delegate");
}

- (void)testCancelConnectionWithResponse_UnimplementedMethodDelegate_NoException {
  MockGHDelegate *delegate = [[MockGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  STAssertNoThrow([abstractResource cancelConnection:mockNSURLConnection() withResponse:mockNSURLResponse()],
                  @"should not have thrown for delegate not implementing githubObject:didCancelFetchWithError:");
}

- (void)testCancelConnectionWithResponse_Connection_CanceledConnection {
  id connection = [OCMockObject mockForClass:[NSURLConnection class]];
  [[connection expect] cancel];
  [abstractResource cancelConnection:connection withResponse:mockNSURLResponse()];
  STAssertNoThrow([connection verify], @"should have called cancel on the connection");
}

- (void)testCancelConnectionWithResponse_NilResponse_NotifiedDelegateFetchCanceled {
  MockCancelGHDelegate *delegate = [[MockCancelGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  [abstractResource cancelConnection:mockNSURLConnection() withResponse:nil];
  STAssertTrue(delegate.called, @"should have called githubObject:didCancelFetchWithError: on delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHConnectionUnknownError, [delegate.error code], @"should have indicated an unknown error");
  NSDictionary *userInfo = [delegate.error userInfo];
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.description", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(2U, [recoveryOptions count], @"should have included two recovery options");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.try_again", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Try the request again") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct first recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.go_back", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Go back and try something different") isEqualToString:[recoveryOptions objectAtIndex:1]],
               @"should have included the correct second recovery option");
}

- (void)testCancelConnectionWithResponse_Response_NotifiedDelegateFetchCanceled {
  MockCancelGHDelegate *delegate = [[MockCancelGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  [abstractResource cancelConnection:mockNSURLConnection() withResponse:mockNSURLResponse()];
  STAssertTrue(delegate.called, @"should have called githubObject:didCancelFetchWithError: on delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHConnectionUnknownError, [delegate.error code], @"should have indicated an unknown error");
  NSDictionary *userInfo = [delegate.error userInfo];
  // TODO: Insert response content somehow into description?
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.description", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(2U, [recoveryOptions count], @"should have included two recovery options");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.try_again", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Try the request again") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct first recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.connection.unknown.recovery_options.go_back", GITHUB_STRINGS_TABLE, @"GHConnectionUnknownError NSLocalizedRecoveryOptionsErrorKey - Go back and try something different") isEqualToString:[recoveryOptions objectAtIndex:1]],
               @"should have included the correct second recovery option");
}

# pragma mark - Github interaction

/** apiPath tests */

- (void)testApiPath_Nil_RootPath {
  STAssertTrue([@"/" isEqualToString:[abstractResource apiPath]], @"should use the root path");
}

/** fetchDataAtApiPath:withParameters: tests */

- (void)testFetchDataAtApiPathWithParameters_NilPath_NoConnection {
  GHAbstractResourceTester *resourceTester = [[GHAbstractResourceTester alloc] init];
  [resourceTester fetchDataAtApiPath:nil withParameters:nil];
  STAssertNil([resourceTester valueForKey:@"_githubConnection"], @"should not have created a Github connection");
}

- (void)testFetchDataAtApiPathWithParameters_PathNoParameters_ConnectionToPath {
  GHAbstractResourceTester *resourceTester = [[GHAbstractResourceTester alloc] init];
  [resourceTester fetchDataAtApiPath:@"/test" withParameters:nil];
  STAssertNotNil([resourceTester valueForKey:@"_githubConnection"], @"should have created a Github connection");
  STAssertTrue([[resourceTester.fetchRequest.URL absoluteString] isEqualToString:@"https://api.github.com/test"],
               @"should have requested the correct URL");
}

- (void)testFetchDataAtApiPathWithParameters_PathWithParameters_ConnectionToPathWithParameters {
  NSDictionary *params = [NSDictionary dictionaryWithObjectsAndKeys:@"test", @"blah", @"test2", @"another", nil];
  GHAbstractResourceTester *resourceTester = [[GHAbstractResourceTester alloc] init];
  [resourceTester fetchDataAtApiPath:@"/test" withParameters:params];
  STAssertNotNil([resourceTester valueForKey:@"_githubConnection"], @"should have created a Github connection");
  STAssertTrue([[resourceTester.fetchRequest.URL absoluteString] isEqualToString:@"https://api.github.com/test?blah=test&another=test2"],
               @"should have requested the correct URL");
}

/** fetchDataWithParameters: tests */

- (void)testFetchDataWithParameters_FetchDataAtApiPathWithParameters {
  id mockResource = [OCMockObject partialMockForObject:abstractResource];
  NSDictionary *parameters = [NSDictionary dictionary];
  [[mockResource expect] fetchDataAtApiPath:[abstractResource apiPath] withParameters:parameters];
  [mockResource fetchDataWithParameters:parameters];
  STAssertNoThrow([mockResource verify], @"should have called fetchDataAtApiPath:apiPath withParameters:parameters");
}

/** fetchData tests */

- (void)testFetchData_FetchDataWithNoParameters {
  id mockResource = [OCMockObject partialMockForObject:abstractResource];
  [[mockResource expect] fetchDataWithParameters:nil];
  [mockResource fetchData];
  STAssertNoThrow([mockResource verify], @"should have called fetchDataWithParameters:nil");
}

/** apiURL tests */

- (void)testApiURL_Path_GithubBaseURLWithPath {
  NSString *path = @"/test";
  id mockResource = [OCMockObject partialMockForObject:abstractResource];
  [[[mockResource stub] andReturn:path] apiPath];
  NSString *URL = [NSString stringWithFormat:@"https://api.github.com%@", path];
  STAssertTrue([URL isEqualToString:[mockResource apiURL]], @"should have appended the path to the Github API base URL");
}

# pragma mark - NSURLConnection delegate methods

/** connection:didReceiveResponse: tests */

- (void)testConnectionDidReceiveResponse_NilConnectionAndResponse_NoException {
  STAssertNoThrow([abstractResource connection:nil didReceiveResponse:nil],
                  @"should not have thrown for nil connection and response");
}

- (void)testConnectionDidReceiveResponse_NilResponse_ExistingJSONParser {
  id jsonParser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [abstractResource setValue:jsonParser forKey:@"_jsonParser"];
  [abstractResource connection:mockNSURLConnection() didReceiveResponse:nil];
  STAssertEquals(jsonParser, [abstractResource valueForKey:@"_jsonParser"],
                 @"should not have destroyed the JSON parser");
}

- (void)testConnectionDidReceiveResponse_NonHTTPResponse_ConnectionCanceledDelegateNotified {
  id mockResource = [OCMockObject partialMockForObject:abstractResource];
  [[mockResource expect] cancelConnection:[OCMArg any] withResponse:[OCMArg any]];
  [mockResource connection:mockNSURLConnection() didReceiveResponse:mockNSURLResponse()];
  STAssertNoThrow([mockResource verify], @"should have canceled the connection");
}

- (void)testConnectionDidReceiveResponse_NonHTTPResponse_NilJSONParser {
  id jsonParser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [abstractResource setValue:jsonParser forKey:@"_jsonParser"];
  [abstractResource connection:mockNSURLConnection() didReceiveResponse:mockNSURLResponse()];
  STAssertNil([abstractResource valueForKey:@"_jsonParser"], @"should have destroyed the JSON parser");
}

- (void)testConnectionDidReceiveResponse_NonSuccessHTTPResponse_ConnectionCanceledDelegateNotified {
  id mockResource = [OCMockObject partialMockForObject:abstractResource];
  [[mockResource expect] cancelConnection:[OCMArg any] withHTTPResponse:[OCMArg any]];
  [abstractResource connection:mockNSURLConnection() didReceiveResponse:mockNSHTTPURLResponse(304)];
  STAssertNoThrow([mockResource verify], @"should have canceled the connection");
}

- (void)testConnectionDidReceiveResponse_NonSuccessHTTPResponse_NilJSONParser {
  id jsonParser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [abstractResource setValue:jsonParser forKey:@"_jsonParser"];
  [abstractResource connection:mockNSURLConnection() didReceiveResponse:mockNSHTTPURLResponse(304)];
  STAssertNil([abstractResource valueForKey:@"_jsonParser"], @"should have destroyed the JSON parser");
}

- (void)testConnectionDidReceiveResponse_SuccessfulResponse_NewJSONParser {
  id oldParser = [abstractResource valueForKey:@"_jsonParser"];
  [abstractResource connection:mockNSURLConnection() didReceiveResponse:mockNSHTTPURLResponse(200)];
  id newParser = [abstractResource valueForKey:@"_jsonParser"];
  STAssertFalse(oldParser == newParser, @"should have created a new JSON parser");
}

/** connection:didReceiveData: tests */

- (void)testConnectionDidReceiveData_NilConnectionAndData_NoException {
  STAssertNoThrow([abstractResource connection:nil didReceiveData:nil],
                  @"should not have thrown for nil connection and data");
}

- (void)testConnectionDidReceiveData_NilParser_NoException {
  [abstractResource setValue:nil forKey:@"_jsonParser"];
  STAssertNoThrow([abstractResource connection:nil didReceiveData:[OCMockObject mockForClass:[NSData class]]],
                  @"should not have thrown for a nil JSON parser");
}

- (void)testConnectionDidReceiveData_Data_ParsedData {
  id data = [OCMockObject mockForClass:[NSData class]];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [[parser expect] parse:data];
  [abstractResource setValue:parser forKey:@"_jsonParser"];
  [abstractResource connection:mockNSURLConnection() didReceiveData:data];
  STAssertNoThrow([parser verify], @"should have parsed the data");
}

/** connection:didFailWithError: tests */

- (void)testConnectionDidFailWithError_NilDelegate_NoException {
  abstractResource.delegate = nil;
  STAssertNoThrow([abstractResource connection:mockNSURLConnection() didFailWithError:mockNSError(@"test")],
                  @"should not have thrown for nil delegate");
}

- (void)testConnectionDidFailWithError_UnimplementedMethodDelegate_NoException {
  MockGHDelegate *delegate = [[MockGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  STAssertNoThrow([abstractResource connection:mockNSURLConnection() didFailWithError:mockNSError(@"test")],
                  @"should not have thrown for delegate not implementing githubObject:didFailWithError:");
}

- (void)testConnectionDidFailWithError_NilConnectionAndError_NoException {
  STAssertNoThrow([abstractResource connection:nil didFailWithError:nil],
                  @"should not have thrown for nil connection and error");
}

- (void)testConnectionDidFailWithError_DestroyParserAndConnection {
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [abstractResource setValue:parser forKey:@"_jsonParser"];
  id connection = mockNSURLConnection();
  [abstractResource setValue:connection forKey:@"_githubConnection"];
  [abstractResource connection:connection didFailWithError:mockNSError(@"test")];
  STAssertNil([abstractResource valueForKey:@"_jsonParser"], @"should have destroyed the JSON parser");
  STAssertNil([abstractResource valueForKey:@"_githubConnection"], @"should have destroyed the Github connection");
}

- (void)testConnectionDidFailWithError_NotifiedDelegate {
  MockFailureGHDelegate *delegate = [[MockFailureGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  NSError *error = mockNSError(@"test");
  [abstractResource connection:mockNSURLConnection() didFailWithError:error];
  STAssertTrue(delegate.called, @"should have called githubObject:connectionDidFailWithError: on delegate");
  STAssertEquals([error userInfo], [delegate.error userInfo], @"should have copied the error to the delegate");
}

/** connectionDidFinishLoading: tests */

- (void)testConnectionDidFinishLoading_NilConnection_NoException {
  STAssertNoThrow([abstractResource connectionDidFinishLoading:nil], @"should not have thrown for nil connection");
}

- (void)testConnectionDidFinishLoading_DestroyParserAndConnection {
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [abstractResource setValue:parser forKey:@"_jsonParser"];
  id connection = mockNSURLConnection();
  [abstractResource setValue:connection forKey:@"_githubConnection"];
  [abstractResource connectionDidFinishLoading:connection];
  STAssertNil([abstractResource valueForKey:@"_jsonParser"], @"should have destroyed the JSON parser");
  STAssertNil([abstractResource valueForKey:@"_githubConnection"], @"should have destroyed the Github connection");
}

- (void)testConnectionDidFinishLoading_NotifiedDelegate {
  MockFinishedGHDelegate *delegate = [[MockFinishedGHDelegate alloc] init];
  abstractResource.delegate = delegate;
  [abstractResource connectionDidFinishLoading:mockNSURLConnection()];
  STAssertTrue(delegate.called, @"should have called githubObjectConnectionDidFinish: on delegate");
}
                   
# pragma mark - Helpers

static id mockNSError(NSString *localizedDescription) {
  return [NSError errorWithDomain:@"com.adamstegman.git-push.GHAbstractResourceTest"
                             code:0
                         userInfo:[NSDictionary dictionaryWithObject:localizedDescription forKey:NSLocalizedDescriptionKey]];
}

static id mockNSURLResponse() {
  id response = [OCMockObject mockForClass:[NSURLResponse class]];
  BOOL notKindOfClass = NO;
  [[[response stub] andReturnValue:OCMOCK_VALUE(notKindOfClass)] isKindOfClass:[OCMArg any]];
  [[[response stub] andReturn:@"application/vnd.git-push"] MIMEType];
  return response;
}

@end
