#import "GHBlobTests.h"
#import "SBJson.h"
#import <OCMock/OCMock.h>
#import "NSString+Base64.h"

/**
 * Subclass to enable testing of the class constructors.
 */
@interface GHBlobTester : GHBlob
@property (nonatomic, strong) NSNumber *fetchingData;
@property (nonatomic, strong) NSURLRequest *fetchRequest;
- (void)connection:(NSURLConnection *)connection didReceiveRequest:(NSURLRequest *)request;
@end
@implementation GHBlobTester
@synthesize fetchingData = _fetchingData;
@synthesize fetchRequest = _fetchRequest;
- (void)fetchData {
  _fetchingData = [NSNumber numberWithBool:YES];
  [super fetchData];
}
- (void)connection:(NSURLConnection *)connection didReceiveRequest:(NSURLRequest *)request {
  self.fetchRequest = request;
}
@end

@implementation GHBlobTests

- (void)setUp {
  [super setUp];
  blob = [[GHBlob alloc] initWithRepository:mockGHRepository(nil)
                              content:[NSData dataWithBytes:[@"test" cStringUsingEncoding:[NSString defaultCStringEncoding]] length:4]
                             encoding:[NSString defaultCStringEncoding]
                                 mode:@"100644"
                                 path:@"test"
                                  sha:@"test"
                                 size:4];
}

- (void)tearDown {
  blob = nil;
  [super tearDown];
}

# pragma mark - Attributes

/** content tests */

- (void)testContent_NSData {
  [blob setValue:[NSData data] forKey:@"_content"];
  STAssertTrue([[NSData data] isEqualToData:blob.content], @"should have stored content data");
}

- (void)testSetContent_NoMethod {
  STAssertThrowsSpecificNamed([blob performSelector:@selector(setContent:) withObject:[NSData data]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setContent: selector");
}

/** encoding tests */

- (void)testEncoding_NSStringEncoding {
  [blob setValue:[NSNumber numberWithInteger:NSProprietaryStringEncoding] forKey:@"_encoding"];
  STAssertEquals((NSStringEncoding)NSProprietaryStringEncoding, blob.encoding, @"should have stored encoding");
}

- (void)testSetEncoding_NoMethod {
  STAssertThrowsSpecificNamed([blob performSelector:@selector(setEncoding:)
                                         withObject:[NSNumber numberWithInteger:NSASCIIStringEncoding]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setEncoding: selector");
}

/** type tests */

- (void)testType_Blob {
  STAssertTrue([@"blob" isEqualToString:blob.type], @"type should be blob");
}

/** contentAsNSString tests */

- (void)testContentAsNSString_NilContent_Nil {
  [blob setValue:nil forKey:@"_content"];
  STAssertNil([blob contentAsNSString], @"should be nil for nil content");
}

- (void)testContentAsNSString_Encoding_StringUsingEncoding {
  [blob setValue:[NSNumber numberWithInteger:NSUTF8StringEncoding] forKey:@"_encoding"];
  NSString *contentString = [NSString stringWithFormat:@"blah %C%C%C blah", 0xA3, 0x0904, 0x1010D];
  NSData *content = [NSData dataWithBytes:[contentString cStringUsingEncoding:NSUTF8StringEncoding]
                                   length:[contentString lengthOfBytesUsingEncoding:NSUTF8StringEncoding]];
  [blob setValue:content forKey:@"_content"];
  STAssertTrue([[blob contentAsNSString] isEqualToString:contentString], @"should have used the stored string encoding");
}

- (void)testContentAsNSString_ProprietaryEncoding_StringUsingDefaultEncoding {
  [blob setValue:[NSNumber numberWithInteger:NSProprietaryStringEncoding] forKey:@"_encoding"];
  NSString *contentString = @"testing";
  NSData *content = [NSData dataWithBytes:[contentString cStringUsingEncoding:[NSString defaultCStringEncoding]]
                                   length:[contentString lengthOfBytesUsingEncoding:[NSString defaultCStringEncoding]]];
  [blob setValue:content forKey:@"_content"];
  STAssertTrue([[blob contentAsNSString] isEqualToString:contentString], @"should have used the default string encoding");
}

/** isText tests */

- (void)testIsText_TextEncoding_Yes {
  [blob setValue:[NSNumber numberWithInteger:NSASCIIStringEncoding] forKey:@"_encoding"];
  STAssertTrue([blob isText], @"should return YES for a text encoding");
}

- (void)testIsText_NSProprietaryStringEncoding_No {
  [blob setValue:[NSNumber numberWithInteger:NSProprietaryStringEncoding] forKey:@"_encoding"];
  STAssertFalse([blob isText], @"should return NO for NSProprietaryStringEncoding");
}

- (void)testIsText_0Encoding_No {
  [blob setValue:[NSNumber numberWithInteger:0] forKey:@"_encoding"];
  STAssertFalse([blob isText], @"should return NO for NSProprietaryStringEncoding");
}

# pragma mark - Class constructors

/** fetchBlobContentWithRepository:sha:delegate: tests */

- (void)testFetchBlobContentWithRepositoryShaDelegate_NilAttributes_NewBlobWithNilAttributes {
  blob = [GHBlob fetchBlobContentWithRepository:nil sha:nil delegate:nil];
  STAssertNil(blob.repository, @"should not have assigned to repository");
  STAssertNil(blob.content, @"should not have assigned content");
  STAssertEquals((NSStringEncoding)NSASCIIStringEncoding, blob.encoding, @"should have defaulted to ASCII encoding");
  STAssertNil(blob.mode, @"should not have assigned to mode");
  STAssertNil(blob.path, @"should not have assigned to path");
  STAssertNil(blob.sha, @"should not have assigned to sha");
  STAssertEquals(0U, blob.size, @"should have assigned size of 0");
  STAssertTrue([@"blob" isEqualToString:blob.type], @"should have assigned blob to type");
  STAssertNil(blob.delegate, @"should not have assigned delegate");
}

- (void)testFetchBlobContentWithRepositoryShaDelegate_Attributes_NewBlobWithAttributes {
  id repository = mockGHRepository(nil);
  NSString *sha = @"test";
  MockGHDelegate *delegate = [[MockGHDelegate alloc] init];
  blob = [GHBlob fetchBlobContentWithRepository:repository sha:sha delegate:delegate];
  STAssertEquals(repository, blob.repository, @"should have assigned repository");
  STAssertNil(blob.content, @"should not have assigned content");
  STAssertEquals((NSStringEncoding)NSASCIIStringEncoding, blob.encoding, @"should have defaulted to ASCII encoding");
  STAssertTrue([sha isEqualToString:blob.sha], @"should have copied sha");
  STAssertEquals(0U, blob.size, @"should not have assigned size");
  STAssertTrue([@"blob" isEqualToString:blob.type], @"should have assigned blob to type");
  STAssertEquals(delegate, blob.delegate, @"should have assigned delegate");
}

- (void)testFetchBlobContentWithRepositoryShaDelegate_StartedFetch {
  GHBlobTester *blobTester = [GHBlobTester fetchBlobContentWithRepository:mockGHRepository(nil) sha:@"test" delegate:nil];
  STAssertTrue([blobTester.fetchingData boolValue], @"should have called #fetchData on the blob");
}

# pragma mark - Initializers

/** initWithRepository:sha: tests */

- (void)testInitWithRepositorySha_NilAttributes_NewBlobWithNilAttributes {
  blob = [blob initWithRepository:nil sha:nil];
  STAssertNil(blob.repository, @"should not have assigned to repository");
  STAssertNil(blob.content, @"should not have assigned content");
  STAssertEquals((NSStringEncoding)NSASCIIStringEncoding, blob.encoding, @"should have defaulted to ASCII encoding");
  STAssertNil(blob.mode, @"should not have assigned mode");
  STAssertNil(blob.path, @"should not have assigned path");
  STAssertNil(blob.sha, @"should not have assigned to sha");
  STAssertEquals(0U, blob.size, @"should have assigned size of 0");
  STAssertTrue([@"blob" isEqualToString:blob.type], @"should have assigned blob to type");
  STAssertNil(blob.delegate, @"should not have assigned delegate");
}

- (void)testInitWithRepositorySha_Attributes_NewBlobWithCorrectAttributes {
  id repository = mockGHRepository(nil);
  NSMutableString *sha = [NSMutableString stringWithString:@"test"];
  blob = [blob initWithRepository:repository sha:sha];
  STAssertEquals(repository, blob.repository, @"should have assigned repository");
  STAssertNil(blob.content, @"should not have assigned content");
  STAssertEquals((NSStringEncoding)NSASCIIStringEncoding, blob.encoding, @"should have defaulted to ASCII encoding");
  STAssertNil(blob.mode, @"should not have assigned mode");
  STAssertNil(blob.path, @"should not have assigned path");
  NSString *oldSha = [sha copy];
  [sha appendString:@"ing"]; // ensure an immutable copy is stored
  STAssertTrue([oldSha isEqualToString:blob.sha], @"should have copied sha");
  STAssertEquals(0U, blob.size, @"should have assigned size of 0");
  STAssertTrue([@"blob" isEqualToString:blob.type], @"should have assigned blob to type");
  STAssertNil(blob.delegate, @"should not have assigned delegate");
}

/** initWithRepository:mode:path:sha:size: tests */

- (void)testInitWithRepositoryModePathShaSize_NilAttributes_NewBlobWithNilAttributes {
  blob = [blob initWithRepository:nil mode:nil path:nil sha:nil size:0];
  STAssertNil(blob.repository, @"should not have assigned to repository");
  STAssertNil(blob.content, @"should not have assigned content");
  STAssertEquals((NSStringEncoding)NSASCIIStringEncoding, blob.encoding, @"should have defaulted to ASCII encoding");
  STAssertNil(blob.mode, @"should not have assigned mode");
  STAssertNil(blob.path, @"should not have assigned path");
  STAssertNil(blob.sha, @"should not have assigned to sha");
  STAssertEquals(0U, blob.size, @"should have assigned size of 0");
  STAssertTrue([@"blob" isEqualToString:blob.type], @"should have assigned blob to type");
  STAssertNil(blob.delegate, @"should not have assigned delegate");
}

- (void)testInitWithRepositoryModePathShaSize_Attributes_NewBlobWithCorrectAttributes {
  id repository = mockGHRepository(nil);
  NSMutableString *mode = [NSMutableString stringWithString:@"100644"];
  NSMutableString *path = [NSMutableString stringWithString:@"testing"];
  NSMutableString *sha = [NSMutableString stringWithString:@"test"];
  NSUInteger size = 15;
  blob = [blob initWithRepository:repository mode:mode path:path sha:sha size:size];
  STAssertEquals(repository, blob.repository, @"should have assigned repository");
  STAssertNil(blob.content, @"should not have assigned content");
  STAssertEquals((NSStringEncoding)NSASCIIStringEncoding, blob.encoding, @"should have defaulted to ASCII encoding");
  NSString *oldMode = [mode copy];
  [mode appendString:@"0"]; // ensure an immutable copy is stored
  STAssertTrue([oldMode isEqualToString:blob.mode], @"should have copied mode");
  NSString *oldPath = [path copy];
  [path appendString:@"/"]; // ensure an immutable copy is stored
  STAssertTrue([oldPath isEqualToString:blob.path], @"should have copied path");
  NSString *oldSha = [sha copy];
  [sha appendString:@"ing"]; // ensure an immutable copy is stored
  STAssertTrue([oldSha isEqualToString:blob.sha], @"should have copied sha");
  STAssertEquals(size, blob.size, @"should have assigned size");
  STAssertTrue([@"blob" isEqualToString:blob.type], @"should have assigned blob to type");
  STAssertNil(blob.delegate, @"should not have assigned delegate");
}

/** initWithRepository:content:encoding:mode:path:sha:size: tests */

- (void)testInitWithRepositoryContentEncodingModePathShaSize_NilAttributes_NewBlobWithNilAttributes {
  blob = [blob initWithRepository:nil content:nil encoding:NSASCIIStringEncoding mode:nil path:nil sha:nil size:0];
  STAssertNil(blob.repository, @"should not have assigned to repository");
  STAssertNil(blob.content, @"should not have assigned content");
  STAssertEquals((NSStringEncoding)NSASCIIStringEncoding, blob.encoding, @"should have assigned encoding");
  STAssertNil(blob.mode, @"should not have assigned mode");
  STAssertNil(blob.path, @"should not have assigned path");
  STAssertNil(blob.sha, @"should not have assigned to sha");
  STAssertEquals(0U, blob.size, @"should have assigned size of 0");
  STAssertTrue([@"blob" isEqualToString:blob.type], @"should have assigned blob to type");
  STAssertNil(blob.delegate, @"should not have assigned delegate");
}

- (void)testInitWithRepositoryContentEncodingModePathShaSize_Attributes_NewBlobWithAttributes {
  id repository = mockGHRepository(nil);
  NSMutableData *content = [NSMutableData data];
  NSMutableString *mode = [NSMutableString stringWithString:@"100644"];
  NSMutableString *path = [NSMutableString stringWithString:@"testing"];
  NSMutableString *sha = [NSMutableString stringWithString:@"test"];
  NSUInteger size = 1;
  blob = [blob initWithRepository:repository content:content encoding:NSUTF8StringEncoding mode:mode path:path sha:sha size:size];
  STAssertEquals(repository, blob.repository, @"should have assigned repository");
  NSData *oldContent = [content copy];
  [content appendBytes:[@"test" cStringUsingEncoding:NSUTF8StringEncoding] length:4]; // ensure an immutable copy is stored
  STAssertTrue([oldContent isEqualToData:blob.content], @"should have copied content");
  STAssertEquals((NSStringEncoding)NSUTF8StringEncoding, blob.encoding, @"should have defaulted to ASCII encoding");
  NSString *oldMode = [mode copy];
  [mode appendString:@"0"]; // ensure an immutable copy is stored
  STAssertTrue([oldMode isEqualToString:blob.mode], @"should have copied mode");
  NSString *oldPath = [path copy];
  [path appendString:@"/"]; // ensure an immutable copy is stored
  STAssertTrue([oldPath isEqualToString:blob.path], @"should have copied path");
  NSString *oldSha = [sha copy];
  [sha appendString:@"ing"]; // ensure an immutable copy is stored
  STAssertTrue([oldSha isEqualToString:blob.sha], @"should have copied sha");
  STAssertEquals(size, blob.size, @"should have assigned size");
  STAssertTrue([@"blob" isEqualToString:blob.type], @"should have assigned blob to type");
  STAssertNil(blob.delegate, @"should not have assigned delegate");
}

# pragma mark - Github interaction

/** apiPath tests */

- (void)testApiPath_BlobPath {
  NSString *owner = @"tester";
  NSString *repo = @"testing";
  GHRepository *repository = [[GHRepository alloc] initWithOwner:owner name:repo];
  NSString *sha = @"test";
  blob = [blob initWithRepository:repository sha:sha];
  NSString *path = [NSString stringWithFormat:@"/repos/%@/%@/git/blobs/%@", owner, repo, sha];
  STAssertTrue([path isEqualToString:[blob apiPath]], @"should return the path to the blob");
}

/** fetchContent tests */

- (void)testFetchContent_FetchData {
  id mockBlob = [OCMockObject partialMockForObject:blob];
  [[mockBlob expect] fetchDataWithParameters:nil];
  [mockBlob fetchContent];
  STAssertNoThrow([mockBlob verify], @"should have called fetchDataWithParameters:");
}

# pragma mark - NSCopying

- (void)testCopying_CopyWithZone_SameObjects {
  GHBlob *newBlob = [blob copyWithZone:NULL];
  STAssertEquals(newBlob.repository, blob.repository, @"should have assigned repository in the new copy");
  STAssertEquals(newBlob.content, blob.content, @"should have assigned content in the new copy");
  STAssertEquals(newBlob.encoding, blob.encoding, @"should have assigned encoding in the new copy");
  STAssertEquals(newBlob.mode, blob.mode, @"should have assigned mode in the new copy");
  STAssertEquals(newBlob.path, blob.path, @"should have assigned path in the new copy");
  STAssertEquals(newBlob.sha, blob.sha, @"should have assigned sha in the new copy");
  STAssertEquals(newBlob.size, blob.size, @"should have assigned size in the new copy");
}

- (void)testCopying_Copy_SameObjects {
  GHBlob *newBlob = [blob copy];
  STAssertEquals(newBlob.repository, blob.repository, @"should have assigned repository in the new copy");
  STAssertEquals(newBlob.content, blob.content, @"should have assigned content in the new copy");
  STAssertEquals(newBlob.encoding, blob.encoding, @"should have assigned encoding in the new copy");
  STAssertEquals(newBlob.mode, blob.mode, @"should have assigned mode in the new copy");
  STAssertEquals(newBlob.path, blob.path, @"should have assigned path in the new copy");
  STAssertEquals(newBlob.sha, blob.sha, @"should have assigned sha in the new copy");
  STAssertEquals(newBlob.size, blob.size, @"should have assigned size in the new copy");
}

# pragma mark - SBJsonStreamParserAdapter delegate methods

/** parser:foundObject: tests */

- (void)testParserFoundObject_Nil_UnchangedAttributes {
  NSData *content = [NSData dataWithBytes:[@"t" cStringUsingEncoding:[NSString defaultCStringEncoding]] length:1];
  [blob setValue:content forKey:@"_content"];
  [blob setValue:[NSNumber numberWithInteger:NSUTF16StringEncoding] forKey:@"_encoding"];
  [blob parser:nil foundObject:nil];
  STAssertTrue([blob.content isEqualToData:content], @"should not have changed content");
  STAssertEquals((NSStringEncoding)NSUTF16StringEncoding, blob.encoding, @"should not have changed encoding");
}

- (void)testParserFoundObject_DifferentKeys_UnchangedAttributes {
  NSData *content = [NSData dataWithBytes:[@"t" cStringUsingEncoding:[NSString defaultCStringEncoding]] length:1];
  [blob setValue:content forKey:@"_content"];
  [blob setValue:[NSNumber numberWithInteger:NSUTF16StringEncoding] forKey:@"_encoding"];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:@"test", @"content",
                                                                    @"utf-8", @"character_encoding", nil];
  [blob parser:parser foundObject:object];
  STAssertTrue([blob.content isEqualToData:content], @"should not have changed content");
  STAssertEquals((NSStringEncoding)NSUTF16StringEncoding, blob.encoding, @"should not have changed encoding");
}

- (void)testParserFoundObject_UTF8Content_UTF8EncodedContent {
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSString *content = [NSString stringWithFormat:@"blah %C%C%C blah", 0xA3, 0x0904, 0x1010D];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:content, @"content",
                                                                    @"utf-8", @"encoding", nil];
  [blob parser:parser foundObject:object];
  NSData *contentData = [NSData dataWithBytes:[content cStringUsingEncoding:NSUTF8StringEncoding]
                                       length:[content lengthOfBytesUsingEncoding:NSUTF8StringEncoding]];
  STAssertTrue([blob.content isEqualToData:contentData], @"should have stored content correctly");
  STAssertEquals((NSStringEncoding)NSUTF8StringEncoding, blob.encoding, @"should have stored encoding correctly");
}

- (void)testParserFoundObject_Base64Content_UTF8EncodedContent {
  // TODO: this is temporary until I can figure out what encoding the content is actually in
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSString *content = [NSString stringWithFormat:@"blah %C%C%C blah", 0xA3, 0x0904, 0x1010D];
  NSString *contentBase64 = [content encodeBase64];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:contentBase64, @"content",
                                                                    @"base64", @"encoding", nil];
  [blob parser:parser foundObject:object];
  NSData *contentData = [NSData dataWithBytes:[content cStringUsingEncoding:NSUTF8StringEncoding]
                                       length:[content lengthOfBytesUsingEncoding:NSUTF8StringEncoding]];
  NSData *contentBase64Data = [contentBase64 decodeBase64];
  STAssertTrue([blob.content isEqualToData:contentBase64Data], @"should have stored content correctly");
  STAssertTrue([blob.content isEqualToData:contentData], @"should have decoded and stored content correctly");
  STAssertEquals((NSStringEncoding)NSUTF8StringEncoding, blob.encoding, @"should have stored encoding correctly");
}

- (void)testParserFoundObject_ExistingData_UpdatedAttributes {
  NSData *content = [NSData dataWithBytes:[@"t" cStringUsingEncoding:[NSString defaultCStringEncoding]] length:1];
  [blob setValue:content forKey:@"_content"];
  [blob setValue:[NSNumber numberWithInteger:NSASCIIStringEncoding] forKey:@"_encoding"];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:@"test", @"content",
                          @"utf-8", @"encoding", nil];
  [blob parser:parser foundObject:object];
  STAssertTrue([[blob contentAsNSString] isEqualToString:@"test"], @"should have stored content correctly");
  STAssertEquals((NSStringEncoding)NSUTF8StringEncoding, blob.encoding, @"should have stored encoding correctly");
}

- (void)testParserFoundObject_BlobWithNSNullAttributes_CorrectAttributes {
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:[NSNull null], @"content",
                          [NSNull null], @"encoding", nil];
  [blob parser:parser foundObject:object];
  STAssertNil(blob.content, @"should have erased content");
  STAssertEquals((NSStringEncoding)0U, blob.encoding, @"should assigned default encoding");
}

@end
