#import "GHCommitTests.h"
#import "GHRepository.h"
#import <OCMock/OCMock.h>

@implementation GHCommitTests

- (void)setUp {
  [super setUp];
  commit = [[GHCommit alloc] init];
}

- (void)tearDown {
  commit = nil;
  [super tearDown];
}

# pragma mark - Attributes

/** author tests */

- (void)testAuthor_NSDictionary {
  [commit setValue:[NSDictionary dictionary] forKey:@"_author"];
  STAssertTrue([[NSDictionary dictionary] isEqualToDictionary:commit.author], @"should have set author");
}

- (void)testSetAuthor_NoMethod {
  STAssertThrowsSpecificNamed([commit performSelector:@selector(setAuthor:) withObject:[NSDictionary dictionary]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setAuthor: selector");
}

/** committer tests */

- (void)testCommitter_NSDictionary {
  [commit setValue:[NSDictionary dictionary] forKey:@"_committer"];
  STAssertTrue([[NSDictionary dictionary] isEqualToDictionary:commit.committer], @"should have set committer");
}

- (void)testSetCommitter_NoMethod {
  STAssertThrowsSpecificNamed([commit performSelector:@selector(setCommitter:) withObject:[NSDictionary dictionary]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setCommitter: selector");
}

/** message tests */

- (void)testMessage_NSString {
  [commit setValue:[NSString string] forKey:@"_message"];
  STAssertTrue([[NSString string] isEqualToString:commit.message], @"should have set message");
}

- (void)testSetMessage_NoMethod {
  STAssertThrowsSpecificNamed([commit performSelector:@selector(setMessage:) withObject:[NSString string]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setMessage: selector");
}

/** parents tests */

- (void)testParents_NSArray {
  [commit setValue:[NSArray array] forKey:@"_parents"];
  STAssertTrue([[NSArray array] isEqualToArray:commit.parents], @"should have set parents");
}

- (void)testSetParents_NoMethod {
  STAssertThrowsSpecificNamed([commit performSelector:@selector(setParents:) withObject:[NSArray array]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setParents: selector");
}

/** tree tests */

- (void)testTree_GHTree {
  GHTree *tree = [[GHTree alloc] init];
  [commit setValue:tree forKey:@"_tree"];
  STAssertEquals(tree, commit.tree, @"should have set tree");
}

- (void)testSetTree_NoMethod {
  STAssertThrowsSpecificNamed([commit performSelector:@selector(setTree:) withObject:[[GHTree alloc] init]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setTree: selector");
}

/** type tests */

- (void)testType_Commit {
  STAssertTrue([@"commit" isEqualToString:commit.type], @"should be a commit type");
}

# pragma mark - Initializers

/** initWithRepository:sha: tests */

- (void)testInitWithRepositorySha_NilAttributes_NewCommitWithNilAttributes {
  commit = [commit initWithRepository:nil sha:nil];
  STAssertNil(commit.repository, @"should not have assigned to repository");
  STAssertNil(commit.sha, @"should not have assigned to sha");
  STAssertNil(commit.delegate, @"should not have assigned delegate");
  STAssertNil(commit.author, @"should not have assigned author");
  STAssertNil(commit.committer, @"should not have assigned committer");
  STAssertNil(commit.message, @"should not have assigned message");
  STAssertNil(commit.parents, @"should not have assigned parents");
  STAssertNil(commit.tree, @"should not have assigned tree");
}

- (void)testInitWithRepositorySha_Attributes_NewCommitWithCorrectAttributes {
  id mockRepository = mockGHRepository(nil);
  commit = [commit initWithRepository:mockRepository sha:@"test"];
  STAssertEquals(mockRepository, commit.repository, @"should have assigned repository");
  STAssertTrue([@"test" isEqualToString:commit.sha], @"should have assigned sha");
  STAssertNil(commit.delegate, @"should not have assigned delegate");
  STAssertNil(commit.author, @"should not have assigned author");
  STAssertNil(commit.committer, @"should not have assigned committer");
  STAssertNil(commit.message, @"should not have assigned message");
  STAssertNil(commit.parents, @"should not have assigned parents");
  STAssertNil(commit.tree, @"should not have assigned tree");
}

/** initWithAttributes: tests */

- (void)testInitWithAttributes_Nil_NilAttributes {
  commit = [commit initWithAttributes:nil];
  STAssertNil(commit.repository, @"should not have assigned to repository");
  STAssertNil(commit.sha, @"should not have assigned to sha");
  STAssertNil(commit.delegate, @"should not have assigned delegate");
  STAssertNil(commit.author, @"should not have assigned author");
  STAssertNil(commit.committer, @"should not have assigned committer");
  STAssertNil(commit.message, @"should not have assigned message");
  STAssertNil(commit.parents, @"should not have assigned parents");
  STAssertNil(commit.tree, @"should not have assigned tree");
}

- (void)testInitWithAttributes_Empty_NilAttributes {
  commit = [commit initWithAttributes:[NSDictionary dictionary]];
  STAssertNil(commit.repository, @"should not have assigned to repository");
  STAssertNil(commit.sha, @"should not have assigned to sha");
  STAssertNil(commit.delegate, @"should not have assigned delegate");
  STAssertNil(commit.author, @"should not have assigned author");
  STAssertNil(commit.committer, @"should not have assigned committer");
  STAssertNil(commit.message, @"should not have assigned message");
  STAssertNil(commit.parents, @"should not have assigned parents");
  STAssertNil(commit.tree, @"should not have assigned tree");
}

- (void)testInitWithAttributes_Attributes_StoredAttributes {
  NSMutableDictionary *attributes = [NSMutableDictionary dictionary];
  id mockRepository = mockGHRepository(nil);
  [attributes setValue:mockRepository forKey:@"repository"];
  [attributes setValue:@"test" forKey:@"sha"];
  NSMutableDictionary *authorAttributes = [NSMutableDictionary dictionary];
  [authorAttributes setValue:@"tester" forKey:@"name"];
  [attributes setValue:authorAttributes forKey:@"author"];
  [attributes setValue:authorAttributes forKey:@"committer"];
  [attributes setValue:@"testing" forKey:@"message"];
  GHCommit *parent = mockGHCommit(nil);
  NSArray *parents = [NSArray arrayWithObject:parent];
  [attributes setValue:parents forKey:@"parents"];
  GHTree *tree = [[GHTree alloc] initWithRepository:mockRepository sha:@"treeSha"];
  [attributes setValue:tree forKey:@"tree"];
  commit = [commit initWithAttributes:attributes];
  STAssertEquals(mockRepository, commit.repository, @"should have assigned repository");
  STAssertTrue([@"test" isEqualToString:commit.sha], @"should have assigned sha");
  STAssertTrue([@"tester" isEqualToString:[commit.author valueForKey:@"name"]], @"should have assigned author");
  STAssertTrue([@"tester" isEqualToString:[commit.committer valueForKey:@"name"]], @"should have assigned committer");
  STAssertTrue([@"testing" isEqualToString:commit.message], @"should have assigned message");
  STAssertEquals(1U, [commit.parents count], @"should have assigned 1 parent");
  STAssertEquals(parent, [commit.parents objectAtIndex:0], @"should have assigned parents");
  STAssertEquals(tree, commit.tree, @"should have assigned tree");
}

- (void)testInitWithAttributes_GithubAPINullAttributes_StoredAttributes {
  NSMutableDictionary *attributes = [NSMutableDictionary dictionary];
  [attributes setValue:[NSNull null] forKey:@"repository"];
  [attributes setValue:[NSNull null] forKey:@"sha"];
  [attributes setValue:[NSNull null] forKey:@"author"];
  [attributes setValue:[NSNull null] forKey:@"committer"];
  [attributes setValue:[NSNull null] forKey:@"message"];
  [attributes setValue:[NSNull null] forKey:@"parents"];
  [attributes setValue:[NSNull null] forKey:@"tree"];
  commit = [commit initWithAttributes:attributes];
  STAssertNil(commit.repository, @"should not have assigned to repository");
  STAssertNil(commit.sha, @"should not have assigned to sha");
  STAssertNil(commit.author, @"should not have assigned author");
  STAssertNil(commit.committer, @"should not have assigned committer");
  STAssertNil(commit.message, @"should not have assigned message");
  STAssertNil(commit.parents, @"should not have assigned parents");
  STAssertNil(commit.tree, @"should not have assigned tree");
}

# pragma mark - Github interaction

/** apiPath tests */

- (void)testApiPath_CommitPath {
  NSString *owner = @"tester";
  NSString *name = @"testing";
  GHRepository *repository = [[GHRepository alloc] initWithOwner:owner name:name];
  NSString *sha = @"test";
  commit = [commit initWithRepository:repository sha:sha];
  NSString *path = [NSString stringWithFormat:@"/repos/%@/%@/git/commits/%@", owner, name, sha];
  STAssertTrue([path isEqualToString:[commit apiPath]], @"should return path to commit");
}

# pragma mark - SBJsonStreamParser delegate methods

/** parser:foundObject: tests */

- (void)testParserFoundObject_Nil_UnchangedAttributes {
  NSMutableDictionary *attributes = [NSMutableDictionary dictionary];
  id mockRepository = mockGHRepository(nil);
  [attributes setValue:mockRepository forKey:@"repository"];
  [attributes setValue:@"test" forKey:@"sha"];
  NSMutableDictionary *authorAttributes = [NSMutableDictionary dictionary];
  [authorAttributes setValue:@"tester" forKey:@"name"];
  [attributes setValue:authorAttributes forKey:@"author"];
  [attributes setValue:authorAttributes forKey:@"committer"];
  [attributes setValue:@"testing" forKey:@"message"];
  GHCommit *parent = mockGHCommit(nil);
  NSArray *parents = [NSArray arrayWithObject:parent];
  [attributes setValue:parents forKey:@"parents"];
  GHTree *tree = [[GHTree alloc] initWithRepository:mockRepository sha:@"treeSha"];
  [attributes setValue:tree forKey:@"tree"];
  commit = [commit initWithAttributes:attributes];
  [commit parser:nil foundObject:nil];
  STAssertEquals(mockRepository, commit.repository, @"should have assigned repository");
  STAssertTrue([@"test" isEqualToString:commit.sha], @"should have assigned sha");
  STAssertTrue([@"tester" isEqualToString:[commit.author valueForKey:@"name"]], @"should have assigned author");
  STAssertTrue([@"tester" isEqualToString:[commit.committer valueForKey:@"name"]], @"should have assigned committer");
  STAssertTrue([@"testing" isEqualToString:commit.message], @"should have assigned message");
  STAssertEquals(1U, [commit.parents count], @"should have assigned 1 parent");
  STAssertEquals(parent, [commit.parents objectAtIndex:0], @"should have assigned parents");
  STAssertEquals(tree, commit.tree, @"should have assigned tree");
}

- (void)testParserFoundObject_DifferentKeys_UnchangedAttributes {
  NSMutableDictionary *attributes = [NSMutableDictionary dictionary];
  id mockRepository = mockGHRepository(nil);
  [attributes setValue:mockRepository forKey:@"repository"];
  [attributes setValue:@"test" forKey:@"sha"];
  NSMutableDictionary *authorAttributes = [NSMutableDictionary dictionary];
  [authorAttributes setValue:@"tester" forKey:@"name"];
  [attributes setValue:authorAttributes forKey:@"author"];
  [attributes setValue:authorAttributes forKey:@"committer"];
  [attributes setValue:@"testing" forKey:@"message"];
  GHCommit *parent = mockGHCommit(nil);
  NSArray *parents = [NSArray arrayWithObject:parent];
  [attributes setValue:parents forKey:@"parents"];
  GHTree *tree = [[GHTree alloc] initWithRepository:mockRepository sha:@"treeSha"];
  [attributes setValue:tree forKey:@"tree"];
  commit = [commit initWithAttributes:attributes];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [commit parser:parser foundObject:[NSDictionary dictionaryWithObject:@"new message" forKey:@"messagee"]];
  STAssertEquals(mockRepository, commit.repository, @"should have assigned repository");
  STAssertTrue([@"test" isEqualToString:commit.sha], @"should have assigned sha");
  STAssertTrue([@"tester" isEqualToString:[commit.author valueForKey:@"name"]], @"should have assigned author");
  STAssertTrue([@"tester" isEqualToString:[commit.committer valueForKey:@"name"]], @"should have assigned committer");
  STAssertTrue([@"testing" isEqualToString:commit.message], @"should have assigned message");
  STAssertEquals(1U, [commit.parents count], @"should have assigned 1 parent");
  STAssertEquals(parent, [commit.parents objectAtIndex:0], @"should have assigned parents");
  STAssertEquals(tree, commit.tree, @"should have assigned tree");
}

- (void)testParserFoundObject_Commit_CorrectAttributes {
  GHRepository *oldRepository = commit.repository;
  NSMutableDictionary *attributes = [NSMutableDictionary dictionary];
  [attributes setValue:@"test" forKey:@"sha"];
  NSMutableDictionary *authorAttributes = [NSMutableDictionary dictionary];
  [authorAttributes setValue:@"tester" forKey:@"name"];
  [attributes setValue:authorAttributes forKey:@"author"];
  [attributes setValue:authorAttributes forKey:@"committer"];
  [attributes setValue:@"testing" forKey:@"message"];
  NSDictionary *parent = [NSDictionary dictionaryWithObjectsAndKeys:@"testParentSha", @"sha", nil];
  [attributes setValue:[NSArray arrayWithObject:parent] forKey:@"parents"];
  NSDictionary *tree = [NSDictionary dictionaryWithObjectsAndKeys:@"testTreeSha", @"sha", nil];
  [attributes setValue:tree forKey:@"tree"];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [commit parser:parser foundObject:attributes];
  STAssertEquals(oldRepository, commit.repository, @"should not have changed repository");
  STAssertTrue([@"test" isEqualToString:commit.sha], @"should have assigned sha");
  STAssertTrue([@"tester" isEqualToString:[commit.author valueForKey:@"name"]], @"should have assigned author");
  STAssertTrue([@"tester" isEqualToString:[commit.committer valueForKey:@"name"]], @"should have assigned committer");
  STAssertTrue([@"testing" isEqualToString:commit.message], @"should have assigned message");
  STAssertEquals(1U, [commit.parents count], @"should have assigned the correct number of parents");
  STAssertTrue([@"testParentSha" isEqualToString:[[commit.parents objectAtIndex:0U] sha]], @"should have assigned parent sha correctly");
  STAssertEquals(oldRepository, [[commit.parents objectAtIndex:0U] repository], @"should have assigned parent's repository correctly");
  STAssertTrue([@"testTreeSha" isEqualToString:[commit.tree sha]], @"should have assigned tree sha correctly");
}

- (void)testParserFoundObject_ExistingData_UpdatedAttributes {
  NSMutableDictionary *oldAttributes = [NSMutableDictionary dictionary];
  id oldRepository = mockGHRepository(nil);
  [oldAttributes setValue:oldRepository forKey:@"repository"];
  [oldAttributes setValue:@"oldSha" forKey:@"sha"];
  NSMutableDictionary *oldAuthorAttributes = [NSMutableDictionary dictionary];
  [oldAuthorAttributes setValue:@"oldAuthor" forKey:@"name"];
  [oldAttributes setValue:oldAuthorAttributes forKey:@"author"];
  [oldAttributes setValue:oldAuthorAttributes forKey:@"committer"];
  [oldAttributes setValue:@"oldMessage" forKey:@"message"];
  GHCommit *oldParent = mockGHCommit(nil);
  NSArray *oldParents = [NSArray arrayWithObject:oldParent];
  [oldAttributes setValue:oldParents forKey:@"parents"];
  GHTree *oldTree = [[GHTree alloc] initWithRepository:oldRepository sha:@"oldTreeSha"];
  [oldAttributes setValue:oldTree forKey:@"tree"];
  commit = [commit initWithAttributes:oldAttributes];
  NSMutableDictionary *attributes = [NSMutableDictionary dictionary];
  [attributes setValue:@"test" forKey:@"sha"];
  NSMutableDictionary *authorAttributes = [NSMutableDictionary dictionary];
  [authorAttributes setValue:@"tester" forKey:@"name"];
  [attributes setValue:authorAttributes forKey:@"author"];
  [attributes setValue:authorAttributes forKey:@"committer"];
  [attributes setValue:@"testing" forKey:@"message"];
  NSDictionary *parent1 = [NSDictionary dictionaryWithObjectsAndKeys:@"testParent1Sha", @"sha", nil];
  NSDictionary *parent2 = [NSDictionary dictionaryWithObjectsAndKeys:@"testParent2Sha", @"sha", nil];
  [attributes setValue:[NSArray arrayWithObjects:parent1, parent2, nil] forKey:@"parents"];
  NSDictionary *tree = [NSDictionary dictionaryWithObjectsAndKeys:@"testTreeSha", @"sha", nil];
  [attributes setValue:tree forKey:@"tree"];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [commit parser:parser foundObject:attributes];
  STAssertEquals(oldRepository, commit.repository, @"should not have changed repository");
  STAssertTrue([@"test" isEqualToString:commit.sha], @"should have assigned sha");
  STAssertTrue([@"tester" isEqualToString:[commit.author valueForKey:@"name"]], @"should have assigned author");
  STAssertTrue([@"tester" isEqualToString:[commit.committer valueForKey:@"name"]], @"should have assigned committer");
  STAssertTrue([@"testing" isEqualToString:commit.message], @"should have assigned message");
  STAssertEquals(2U, [commit.parents count], @"should have assigned the correct number of parents");
  STAssertTrue([@"testParent1Sha" isEqualToString:[[commit.parents objectAtIndex:0U] sha]], @"should have assigned parent 1 sha correctly");
  STAssertEquals(oldRepository, [[commit.parents objectAtIndex:0U] repository], @"should have assigned parent's repository correctly");
  STAssertTrue([@"testParent2Sha" isEqualToString:[[commit.parents objectAtIndex:1U] sha]], @"should have assigned parent 2 sha correctly");
  STAssertEquals(oldRepository, [[commit.parents objectAtIndex:1U] repository], @"should have assigned parent's repository correctly");
  STAssertTrue([@"testTreeSha" isEqualToString:[commit.tree sha]], @"should have assigned tree sha correctly");
}

- (void)testParserFoundObject_CommitWithNSNullAttributes_CorrectAttributes {
  NSMutableDictionary *attributes = [NSMutableDictionary dictionary];
  [attributes setValue:[NSNull null] forKey:@"repository"];
  [attributes setValue:[NSNull null] forKey:@"sha"];
  [attributes setValue:[NSNull null] forKey:@"author"];
  [attributes setValue:[NSNull null] forKey:@"committer"];
  [attributes setValue:[NSNull null] forKey:@"message"];
  [attributes setValue:[NSNull null] forKey:@"parents"];
  [attributes setValue:[NSNull null] forKey:@"tree"];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [commit parser:parser foundObject:attributes];
  STAssertNil(commit.repository, @"should not have assigned to repository");
  STAssertNil(commit.sha, @"should not have assigned to sha");
  STAssertNil(commit.author, @"should not have assigned author");
  STAssertNil(commit.committer, @"should not have assigned committer");
  STAssertNil(commit.message, @"should not have assigned message");
  STAssertNil(commit.parents, @"should not have assigned parents");
  STAssertNil(commit.tree, @"should not have assigned tree");
}

@end
