#import "GHGitObjectTests.h"
#import "GHErrors.h"
#import <OCMock/OCMock.h>

/**
 * A subclass used to ensure the correct request is generated. Stores the request for later retrieval by the test.
 */
@interface GHGitObjectTester : GHGitObject
@property (nonatomic, strong) NSURLRequest *fetchRequest;
- (void)connection:(NSURLConnection *)connection didReceiveRequest:(NSURLRequest *)request;
@end
@implementation GHGitObjectTester
@synthesize fetchRequest = _fetchRequest;
- (void)connection:(NSURLConnection *)connection didReceiveRequest:(NSURLRequest *)request {
  self.fetchRequest = request;
}
@end

@implementation GHGitObjectTests

- (void)setUp {
  gitObject = [[GHGitObject alloc] init];
}

- (void)tearDown {
  gitObject = nil;
}

# pragma mark - Attributes

/** sha tests */

- (void)testSha_NSString {
  [gitObject setValue:[NSString string] forKey:@"_sha"];
  STAssertTrue([[NSString string] isEqualToString:gitObject.sha], @"should have stored sha string");
}

- (void)testSetSha_NoMethod {
  STAssertThrowsSpecificNamed([gitObject performSelector:@selector(setSha:) withObject:[NSString string]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setSha: selector");
}

# pragma mark - Github interaction

/** fetchDataWithParameters tests */

- (void)testFetchDataWithParameters_NilRepository_Failure {
  [gitObject setValue:nil forKey:@"_repository"];
  MockCannotGHDelegate *delegate = [[MockCannotGHDelegate alloc] init];
  gitObject.delegate = delegate;
  [gitObject fetchDataWithParameters:nil];
  STAssertNil([gitObject valueForKey:@"_githubConnection"], @"should not have created a connection");
  STAssertTrue(delegate.called, @"should have called githubObject:cannotFetchContentDueToError: on the delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHNoRepositoryError, [delegate.error code], @"should have indicated a missing repository as the error");
  NSDictionary *userInfo = [delegate.error userInfo];
  STAssertTrue([NSLocalizedStringFromTable(@"github.no_repository_error.description", GITHUB_STRINGS_TABLE, @"GHNoRepositoryError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.no_repository_error.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHNoRepositoryError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(1U, [recoveryOptions count], @"should have included one recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct recovery option");
}

- (void)testFetchDataWithParameters_NilSha_Failure {
  [gitObject setValue:mockGHRepository(nil) forKey:@"_repository"];
  [gitObject setValue:nil forKey:@"_sha"];
  MockCannotGHDelegate *delegate = [[MockCannotGHDelegate alloc] init];
  gitObject.delegate = delegate;
  [gitObject fetchDataWithParameters:nil];
  STAssertNil([gitObject valueForKey:@"_githubConnection"], @"should not have created a connection");
  STAssertTrue(delegate.called, @"should have called githubObject:cannotFetchContentDueToError: on the delegate");
  STAssertTrue([GITHUB_ERROR_DOMAIN isEqualToString:[delegate.error domain]], @"should have used the Github error domain");
  STAssertEquals(GHNoShaError, [delegate.error code], @"should have indicated a missing sha as the error");
  NSDictionary *userInfo = [delegate.error userInfo];
  STAssertTrue([NSLocalizedStringFromTable(@"github.no_sha_error.description", GITHUB_STRINGS_TABLE, @"GHNoShaError NSLocalizedDescriptionKey") isEqualToString:[userInfo objectForKey:NSLocalizedDescriptionKey]],
               @"should have included a description for the error");
  STAssertTrue([NSLocalizedStringFromTable(@"github.no_sha_error.recovery_suggestion", GITHUB_STRINGS_TABLE, @"GHNoShaError NSLocalizedRecoverySuggestionErrorKey") isEqualToString:[userInfo objectForKey:NSLocalizedRecoverySuggestionErrorKey]],
               @"should have included a recovery suggestion for the error");
  NSArray *recoveryOptions = [userInfo objectForKey:NSLocalizedRecoveryOptionsErrorKey];
  STAssertEquals(1U, [recoveryOptions count], @"should have included one recovery option");
  STAssertTrue([NSLocalizedStringFromTable(@"github.common.recovery_options.acknowledgement", GITHUB_STRINGS_TABLE, @"NSLocalizedRecoveryOptionsErrorKey - Acknowledgment (OK); have instructed the user what to do to fix the issue") isEqualToString:[recoveryOptions objectAtIndex:0]],
               @"should have included the correct recovery option");
}

- (void)testFetchDataWithParameters_RepositorySha_Super {
  // mirrors the test for GHAbstractResource since I can't make an expectation on super.
  GHGitObjectTester *resourceTester = [[GHGitObjectTester alloc] init];
  [resourceTester setValue:mockGHRepository(nil) forKey:@"_repository"];
  [resourceTester setValue:@"test" forKey:@"_sha"];
  [resourceTester fetchDataWithParameters:nil];
  STAssertNotNil([resourceTester valueForKey:@"_githubConnection"], @"should have created a Github connection");
  STAssertTrue([[resourceTester.fetchRequest.URL absoluteString] isEqualToString:@"https://api.github.com/"],
               @"should have requested the correct URL");
}

@end
