#import "GHRefTests.h"
#import "GHRepository.h"
#import <OCMock/OCMock.h>

@implementation GHRefTests

- (void)setUp {
  [super setUp];
  ref = [[GHRef alloc] init];
}

- (void)tearDown {
  ref = nil;
  [super tearDown];
}

# pragma mark - Attributes

/** object tests */

- (void)testObject_GHCommit {
  GHCommit *commit = [[GHCommit alloc] init];
  [ref setValue:commit forKey:@"_object"];
  STAssertEquals(commit, ref.object, @"should have set object");
}

- (void)testSetObject_NoMethod {
  STAssertThrowsSpecificNamed([ref performSelector:@selector(setObject:) withObject:[[GHCommit alloc] init]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setObject: selector");
}

/** ref tests */

- (void)testRef_NSString {
  [ref setValue:[NSString string] forKey:@"_ref"];
  STAssertTrue([[NSString string] isEqualToString:ref.ref], @"should have set ref");
}

- (void)testSetRef_NoMethod {
  STAssertThrowsSpecificNamed([ref performSelector:@selector(setRef:) withObject:[NSString string]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setRef: selector");
}

# pragma mark - Initializers

/** initWithRepository:ref: tests */

- (void)testInitWithRepositoryRef_NilAttributes_NewRefWithNilAttributes {
  ref = [ref initWithRepository:nil ref:nil];
  STAssertNil(ref.repository, @"should not have assigned repository");
  STAssertNil(ref.ref, @"should not have assigned ref");
  STAssertNil(ref.object, @"should not have assigned object");
}

- (void)testInitWithRepositoryRef_Attributes_NewRefWithCorrectAttributes {
  id repository = mockGHRepository(nil);
  ref = [ref initWithRepository:repository ref:@"refs/heads/master"];
  STAssertEquals(repository, ref.repository, @"should have assigned repository");
  STAssertTrue([@"refs/heads/master" isEqualToString:ref.ref], @"should have assigned ref");
  STAssertNil(ref.object, @"should not have assigned object");
}

/** initWithRepository:ref:object: tests */

- (void)testInitWithRepositoryRefObject_NilAttributes_NewRefWithNilAttributes {
  ref = [ref initWithRepository:nil ref:nil object:nil];
  STAssertNil(ref.repository, @"should not have assigned repository");
  STAssertNil(ref.ref, @"should not have assigned ref");
  STAssertNil(ref.object, @"should not have assigned object");
}

- (void)testInitWithRepositoryRefObject_Attributes_NewRefWithCorrectAttributes {
  id repository = mockGHRepository(nil);
  id commit = mockGHCommit(nil);
  ref = [ref initWithRepository:repository ref:@"refs/heads/master" object:commit];
  STAssertEquals(repository, ref.repository, @"should have assigned repository");
  STAssertTrue([@"refs/heads/master" isEqualToString:ref.ref], @"should have assigned ref");
  STAssertEquals(commit, ref.object, @"should have assigned object");
}

# pragma mark - Github interaction

- (void)testApiPath_RefPath {
  GHRepository *repository = mockGHRepository(nil);
  NSString *reference = @"refs/heads/master";
  NSString *path = [NSString stringWithFormat:@"/repos/%@/%@/git/%@", repository.owner.login, repository.name, reference];
  ref = [ref initWithRepository:repository ref:reference];
  STAssertTrue([path isEqualToString:[ref apiPath]], @"should have returned path to ref");
}

# pragma mark - SBJsonStreamParser delegate methods

/** parser:foundObject: tests */

- (void)testParserFoundObject_Nil_UnchangedAttributes {
  id repository = mockGHRepository(nil);
  id commit = mockGHCommit(nil);
  ref = [ref initWithRepository:repository ref:@"test" object:commit];
  [ref parser:nil foundObject:nil];
  STAssertEquals(repository, ref.repository, @"should not have changed repository");
  STAssertEquals(commit, ref.object, @"should not have changed object");
  STAssertTrue([@"test" isEqualToString:ref.ref], @"should not have changed ref");
}

- (void)testParserFoundObject_DifferentKeys_UnchangedAttributes {
  id repository = mockGHRepository(nil);
  id commit = mockGHCommit(nil);
  ref = [ref initWithRepository:repository ref:@"test" object:commit];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:@"new ref", @"reff", nil];
  [ref parser:parser foundObject:object];
  STAssertEquals(repository, ref.repository, @"should not have changed repository");
  STAssertEquals(commit, ref.object, @"should not have changed object");
  STAssertTrue([@"test" isEqualToString:ref.ref], @"should not have changed ref");
}

- (void)testParserFoundObject_Ref_CorrectAttributes {
  id repository = mockGHRepository(nil);
  ref = [ref initWithRepository:repository ref:@"doesn't matter"];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSDictionary *commit = [NSDictionary dictionaryWithObjectsAndKeys:@"commit", @"type",
                          @"testSha", @"sha", nil];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:@"test", @"ref",
                          commit, @"object", nil];
  [ref parser:parser foundObject:object];
  STAssertEquals(repository, ref.repository, @"should not have changed repository");
  STAssertEquals(repository, ref.object.repository, @"should have assigned commit repository");
  STAssertTrue([@"testSha" isEqualToString:ref.object.sha], @"should have assigned commit sha");
  STAssertTrue([@"test" isEqualToString:ref.ref], @"should have assigned ref");
}

- (void)testParserFoundObject_ExistingData_UpdatedAttributes {
  id oldRepository = mockGHRepository(nil);
  id oldCommit = mockGHCommit([NSDictionary dictionaryWithObjectsAndKeys:@"oldSha", @"sha", nil]);
  ref = [ref initWithRepository:oldRepository ref:@"oldRef" object:oldCommit];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSDictionary *commit = [NSDictionary dictionaryWithObjectsAndKeys:@"commit", @"type",
                          @"testSha", @"sha", nil];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:@"test", @"ref",
                          commit, @"object", nil];
  [ref parser:parser foundObject:object];
  STAssertEquals(oldRepository, ref.repository, @"should not have changed repository");
  STAssertEquals(oldRepository, ref.object.repository, @"should have assigned commit repository");
  STAssertTrue([@"testSha" isEqualToString:ref.object.sha], @"should have assigned commit sha");
  STAssertTrue([@"test" isEqualToString:ref.ref], @"should have assigned ref");
}

- (void)testParserFoundObject_RefWithNSNullAttributes_CorrectAttributes {
  id oldRepository = mockGHRepository(nil);
  id oldCommit = mockGHCommit([NSDictionary dictionaryWithObjectsAndKeys:@"oldSha", @"sha", nil]);
  ref = [ref initWithRepository:oldRepository ref:@"oldRef" object:oldCommit];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [ref parser:parser foundObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSNull null], @"ref",
                                  [NSNull null], @"object", nil]];
  STAssertNil(ref.ref, @"should not have assigned ref");
  STAssertNil(ref.object, @"should not have assigned object");
}

@end
