#import "GHTreeTests.h"
#import <OCMock/OCMock.h>
#import "GHTreeLeaf.h"
#import "GHBlob.h"
#import "GHErrors.h"

@implementation GHTreeTests

- (void)setUp {
  [super setUp];
  GHTreeLeaf *leaf = [[GHTreeLeaf alloc] init];
  tree = [[GHTree alloc] initWithRepository:mockGHRepository(nil)
                               leaves:[NSArray arrayWithObject:leaf]
                                 mode:@"100777"
                                 path:@"testing"
                                  sha:@"test"];
}

- (void)tearDown {
  tree = nil;
  [super tearDown];
}

# pragma mark - Attributes

/** leaves tests */

- (void)testLeaves_NSArray {
  GHTreeLeaf *leaf = [[GHTreeLeaf alloc] init];
  NSArray *leaves = [NSArray arrayWithObject:leaf];
  [tree setValue:leaves forKey:@"_leaves"];
  STAssertEquals(leaves, [tree leaves], @"should have returned the leaves array");
}

/**
 * Should not allow writes to leaves.
 */
- (void)testSetLeaves_NoMethod {
  STAssertThrowsSpecificNamed([tree performSelector:@selector(setLeaves:) withObject:[NSArray array]],
                              NSException, NSInvalidArgumentException,
                              @"should have thrown invalid argument exception for setLeaves: selector");
}

/** type tests */

- (void)testType_Tree {
  STAssertTrue([@"tree" isEqualToString:tree.type], @"should be a tree type");
}

# pragma mark - Initializers

/** initWithRepository:sha tests */

- (void)testInitWithRepositorySha_NilAttributes_NewTreeWithNilAttributes {
  tree = [tree initWithRepository:nil sha:nil];
  STAssertNil(tree.repository, @"should not have assigned to repository");
  STAssertNil(tree.leaves, @"should not have assigned leaves");
  STAssertNil(tree.mode, @"should not have assigned mode");
  STAssertNil(tree.path, @"should not have assigned path");
  STAssertNil(tree.sha, @"should not have assigned to sha");
  STAssertEquals(0U, tree.size, @"should have assigned size of 0");
  STAssertTrue([@"tree" isEqualToString:tree.type], @"should have assigned tree to type");
  STAssertNil(tree.delegate, @"should not have assigned delegate");
}

- (void)testInitWithRepositorySha_Attributes_NewTreeWithCorrectAttributes {
  id repository = mockGHRepository(nil);
  NSMutableString *sha = [NSMutableString stringWithString:@"test"];
  tree = [tree initWithRepository:repository sha:sha];
  STAssertEquals(repository, tree.repository, @"should have assigned repository");
  STAssertNil(tree.leaves, @"should not have assigned leaves");
  STAssertNil(tree.mode, @"should not have assigned mode");
  STAssertNil(tree.path, @"should not have assigned path");
  NSString *oldSha = [sha copy];
  [sha appendString:@"ing"]; // ensure an immutable copy is stored
  STAssertTrue([oldSha isEqualToString:tree.sha], @"should have copied sha");
  STAssertEquals(0U, tree.size, @"should have assigned size of 0");
  STAssertTrue([@"tree" isEqualToString:tree.type], @"should have assigned tree to type");
  STAssertNil(tree.delegate, @"should not have assigned delegate");
}

/** initWithRepository:mode:path:sha: tests */

- (void)testInitWithRepositoryModePathSha_NilAttributes_NewTreeWithNilAttributes {
  tree = [tree initWithRepository:nil mode:nil path:nil sha:nil];
  STAssertNil(tree.repository, @"should not have assigned to repository");
  STAssertNil(tree.leaves, @"should not have assigned leaves");
  STAssertNil(tree.mode, @"should not have assigned mode");
  STAssertNil(tree.path, @"should not have assigned path");
  STAssertNil(tree.sha, @"should not have assigned to sha");
  STAssertEquals(0U, tree.size, @"should have assigned size of 0");
  STAssertTrue([@"tree" isEqualToString:tree.type], @"should have assigned tree to type");
  STAssertNil(tree.delegate, @"should not have assigned delegate");
}

- (void)testInitWithRepositoryModePathSha_Attributes_NewTreeWithCorrectAttributes {
  id repository = mockGHRepository(nil);
  NSMutableString *mode = [NSMutableString stringWithString:@"100644"];
  NSMutableString *path = [NSMutableString stringWithString:@"testing"];
  NSMutableString *sha = [NSMutableString stringWithString:@"test"];
  tree = [tree initWithRepository:repository mode:mode path:path sha:sha];
  STAssertEquals(repository, tree.repository, @"should have assigned repository");
  STAssertNil(tree.leaves, @"should not have assigned leaves");
  NSString *oldMode = [mode copy];
  [mode appendString:@"0"]; // ensure an immutable copy is stored
  STAssertTrue([oldMode isEqualToString:tree.mode], @"should have copied mode");
  NSString *oldPath = [path copy];
  [path appendString:@"/"]; // ensure an immutable copy is stored
  STAssertTrue([oldPath isEqualToString:tree.path], @"should have copied path");
  NSString *oldSha = [sha copy];
  [sha appendString:@"ing"]; // ensure an immutable copy is stored
  STAssertTrue([oldSha isEqualToString:tree.sha], @"should have copied sha");
  STAssertEquals(0U, tree.size, @"should have assigned size of 0");
  STAssertTrue([@"tree" isEqualToString:tree.type], @"should have assigned tree to type");
  STAssertNil(tree.delegate, @"should not have assigned delegate");
}

/** initWithRepository:leaves:mode:path:sha: tests */

- (void)testInitWithRepositoryLeavesModePathSha_NilAttributes_NewTreeWithNilAttributes {
  tree = [tree initWithRepository:nil leaves:nil mode:nil path:nil sha:nil];
  STAssertNil(tree.repository, @"should not have assigned to repository");
  STAssertNil(tree.leaves, @"should not have assigned leaves");
  STAssertNil(tree.mode, @"should not have assigned mode");
  STAssertNil(tree.path, @"should not have assigned path");
  STAssertNil(tree.sha, @"should not have assigned to sha");
  STAssertEquals(0U, tree.size, @"should have assigned size of 0");
  STAssertTrue([@"tree" isEqualToString:tree.type], @"should have assigned tree to type");
  STAssertNil(tree.delegate, @"should not have assigned delegate");
}

- (void)testInitWithRepositoryLeavesModePathSha_Attributes_NewTreeWithCorrectAttributes {
  id repository = mockGHRepository(nil);
  GHTreeLeaf *leaf = [[GHTreeLeaf alloc] init];
  NSMutableArray *leaves = [NSMutableArray arrayWithObject:leaf];
  NSMutableString *mode = [NSMutableString stringWithString:@"100644"];
  NSMutableString *path = [NSMutableString stringWithString:@"testing"];
  NSMutableString *sha = [NSMutableString stringWithString:@"test"];
  tree = [tree initWithRepository:repository leaves:leaves mode:mode path:path sha:sha];
  STAssertEquals(repository, tree.repository, @"should have assigned repository");
  [leaves addObject:[[GHTreeLeaf alloc] init]]; // ensure the tree copied the leaves instead of assigning
  STAssertEquals(1U, [[tree leaves] count], @"should have stored the leaves array with the correct number of elements");
  STAssertEquals(leaf, [[tree leaves] objectAtIndex:0], @"should have stored the leaves array");
  NSString *oldMode = [mode copy];
  [mode appendString:@"0"]; // ensure an immutable copy is stored
  STAssertTrue([oldMode isEqualToString:tree.mode], @"should have copied mode");
  NSString *oldPath = [path copy];
  [path appendString:@"/"]; // ensure an immutable copy is stored
  STAssertTrue([oldPath isEqualToString:tree.path], @"should have copied path");
  NSString *oldSha = [sha copy];
  [sha appendString:@"ing"]; // ensure an immutable copy is stored
  STAssertTrue([oldSha isEqualToString:tree.sha], @"should have copied sha");
  STAssertEquals(0U, tree.size, @"should have assigned size of 0");
  STAssertTrue([@"tree" isEqualToString:tree.type], @"should have assigned tree to type");
  STAssertNil(tree.delegate, @"should not have assigned delegate");
}

# pragma mark - Github interaction

/** apiPath tests */

- (void)testApiPath_TreePath {
  NSString *owner = @"tester";
  NSString *repo = @"testing";
  GHRepository *repository = [[GHRepository alloc] initWithOwner:owner name:repo];
  NSString *sha = @"test";
  tree = [tree initWithRepository:repository sha:sha];
  NSString *path = [NSString stringWithFormat:@"/repos/%@/%@/git/trees/%@", owner, repo, sha];
  STAssertTrue([path isEqualToString:[tree apiPath]], @"should return the path to the tree");
}

/** fetchImmediateLeaves tests */

- (void)testFetchImmediateLeaves_FetchDataWithNoParameters {
  id mockTree = [OCMockObject partialMockForObject:tree];
  [[mockTree expect] fetchDataWithParameters:nil];
  [mockTree fetchImmediateLeaves];
  STAssertNoThrow([mockTree verify], @"should have called fetchDataWithParameters:nil");
}

/** fetchLeaves tests */

- (void)testFetchLeaves_FetchDataWithRecursiveParameter {
  NSDictionary *params = [NSDictionary dictionaryWithObjectsAndKeys:@"1", @"recursive", nil];
  id mockTree = [OCMockObject partialMockForObject:tree];
  [[mockTree expect] fetchDataWithParameters:params];
  [mockTree fetchLeaves];
  STAssertNoThrow([mockTree verify], @"should have called fetchDataWithParameters:{recursive=>1}");
}

# pragma mark - NSCopying

/** copyWithZone: tests */

- (void)testCopying_CopyWithZone_SameObjects {
  STAssertTrue([[tree leaves] count] > 0U, @"should have leaves in the tree");
  STAssertTrue([tree leaves] == [[tree copyWithZone:NULL] leaves], @"should have assigned the array in the new copy");
}

/** copy tests */

- (void)testCopying_Copy_SameObjects {
  STAssertTrue([[tree leaves] count] > 0U, @"should have leaves in the tree");
  STAssertTrue([tree leaves] == [[tree copy] leaves], @"should have assigned the array in the new copy");
}

# pragma mark - NSFastEnumeration

- (void)testFastEnumeration_ForInLoop_IterateOverLeaves {
  STAssertTrue([[tree leaves] count] > 0U, @"should have leaves in the tree");
  NSMutableArray *leaves = [NSMutableArray array];
  void (^ForInLoop)(void) = ^(void) {
    for (GHTreeLeaf *leaf in tree) {
      [leaves addObject:leaf];
    }
  };
  STAssertNoThrow(ForInLoop(), @"should not throw an exception when performing a for...in loop");
  STAssertEquals([[tree leaves] count], [leaves count], @"should have iterated over the leaves in the tree");
}

# pragma mark - SBJsonStreamParserAdapter delegate methods

/** parser:foundObject: tests */

- (void)testParserFoundObject_Nil_UnchangedAttributes {
  NSArray *oldLeaves = [NSArray arrayWithArray:tree.leaves];
  STAssertTrue([tree.leaves count] > 0U, @"should have leaves in the tree");
  [tree parser:nil foundObject:nil];
  STAssertEquals([oldLeaves count], [tree.leaves count], @"should not have a different number of leaves");
  for (int i = 0, len = [tree.leaves count]; i < len; i++) {
    STAssertEquals([oldLeaves objectAtIndex:i], [tree.leaves objectAtIndex:i], @"should not have changed any leaf");
  }
}

- (void)testParserFoundObject_DifferentKeys_UnchangedAttributes {
  NSArray *oldLeaves = [NSArray arrayWithArray:tree.leaves];
  NSString *oldSha = tree.sha;
  STAssertTrue([tree.leaves count] > 0U, @"should have leaves in the tree");
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:@"test", @"shaa",
                          @"testing", @"urll",
                          [NSArray array], @"treee", nil];
  [tree parser:parser foundObject:object];
  STAssertEquals([oldLeaves count], [tree.leaves count], @"should not have a different number of leaves");
  for (int i = 0, len = [tree.leaves count]; i < len; i++) {
    STAssertEquals([oldLeaves objectAtIndex:i], [tree.leaves objectAtIndex:i], @"should not have changed any leaf");
  }
  STAssertTrue([oldSha isEqualToString:tree.sha], @"should not have changed sha");
}

- (void)testParserFoundObject_Tree_CorrectAttributes {
  tree = [tree initWithRepository:mockGHRepository(nil) leaves:nil mode:nil path:nil sha:nil];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSString *url = [NSString stringWithFormat:@"https://api.github.com/%@/%@/repos/git/trees/%@",
                   tree.repository.owner, tree.repository.name, tree.sha];
  GHTreeLeaf *leaf = [[GHBlob alloc] initWithRepository:mockGHRepository(nil) mode:@"100644" path:@"testing.txt" sha:@"test-file" size:10U];
  NSString *leafUrl = [NSString stringWithFormat:@"https://api.github.com/%@/%@/repos/git/blobs/%@",
                       tree.repository.owner, tree.repository.name, leaf.sha];
  GHTreeLeaf *branch = [[GHTree alloc] initWithRepository:mockGHRepository(nil) mode:@"100644" path:@"testing-dir" sha:@"test-subdir"];
  NSString *branchUrl = [NSString stringWithFormat:@"https://api.github.com/%@/%@/repos/git/trees/%@",
                         tree.repository.owner, tree.repository.name, branch.sha];
  NSDictionary *leafObject = [NSDictionary dictionaryWithObjectsAndKeys:leaf.path, @"path",
                              leaf.mode, @"mode",
                              @"blob", @"type",
                              [NSNumber numberWithInteger:((GHBlob *)leaf).size], @"size",
                              leaf.sha, @"sha",
                              leafUrl, @"url", nil];
  NSDictionary *branchObject = [NSDictionary dictionaryWithObjectsAndKeys:branch.path, @"path",
                                branch.mode, @"mode",
                                @"tree", @"type",
                                branch.sha, @"sha",
                                branchUrl, @"url", nil];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:@"test", @"sha",
                          url, @"url",
                          [NSArray arrayWithObjects:leafObject, branchObject, nil], @"tree", nil];
  [tree parser:parser foundObject:object];
  STAssertEquals(2U, [tree.leaves count], @"should have stored the leaves array with the correct number of elements");
  NSMutableArray *types = [NSMutableArray arrayWithObjects:[GHBlob class], [GHTree class], nil];
  for (GHTreeLeaf *newLeaf in tree) {
    [types removeObject:[newLeaf class]];
    if ([newLeaf isKindOfClass:[GHBlob class]]) {
      STAssertTrue([leaf.mode isEqualToString:newLeaf.mode], @"should have stored mode in the blob correctly");
      STAssertTrue([leaf.path isEqualToString:newLeaf.path], @"should have stored path in the blob correctly");
      STAssertTrue([leaf.sha isEqualToString:newLeaf.sha], @"should have stored sha in the blob correctly");
      STAssertEquals(leaf.size, newLeaf.size, @"should have stored size in the blob correctly");
      STAssertTrue([@"blob" isEqualToString:newLeaf.type], @"should have stored type in the blob correctly");
    } else {
      STAssertTrue([branch.mode isEqualToString:newLeaf.mode], @"should have stored mode in the tree correctly");
      STAssertTrue([branch.path isEqualToString:newLeaf.path], @"should have stored path in the tree correctly");
      STAssertTrue([branch.sha isEqualToString:newLeaf.sha], @"should have stored sha in the tree correctly");
      STAssertEquals(branch.size, newLeaf.size, @"should have stored size in the tree correctly");
      STAssertTrue([@"tree" isEqualToString:newLeaf.type], @"should have stored type in the tree correctly");
    }
  }
  STAssertTrue([types count] == 0, @"should have had a blob and a tree");
  STAssertTrue([@"test" isEqualToString:tree.sha], @"should have updated sha");
}

- (void)testParserFoundObject_ExistingData_UpdatedAttributes {
  GHTreeLeaf *oldLeaf = [[GHTreeLeaf alloc] init];
  tree = [tree initWithRepository:mockGHRepository(nil) leaves:[NSArray arrayWithObject:oldLeaf] mode:@"test" path:@"blah" sha:@"something"];
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  NSString *url = [NSString stringWithFormat:@"https://api.github.com/%@/%@/repos/git/trees/%@",
                   tree.repository.owner, tree.repository.name, tree.sha];
  GHTreeLeaf *leaf = [[GHBlob alloc] initWithRepository:mockGHRepository(nil) mode:@"100644" path:@"testing.txt" sha:@"test-file" size:10U];
  NSString *leafUrl = [NSString stringWithFormat:@"https://api.github.com/%@/%@/repos/git/blobs/%@",
                       tree.repository.owner, tree.repository.name, leaf.sha];
  GHTreeLeaf *branch = [[GHTree alloc] initWithRepository:mockGHRepository(nil) mode:@"100644" path:@"testing-dir" sha:@"test-subdir"];
  NSString *branchUrl = [NSString stringWithFormat:@"https://api.github.com/%@/%@/repos/git/trees/%@",
                         tree.repository.owner, tree.repository.name, branch.sha];
  NSDictionary *leafObject = [NSDictionary dictionaryWithObjectsAndKeys:leaf.path, @"path",
                              leaf.mode, @"mode",
                              @"blob", @"type",
                              [NSNumber numberWithInteger:((GHBlob *)leaf).size], @"size",
                              leaf.sha, @"sha",
                              leafUrl, @"url", nil];
  NSDictionary *branchObject = [NSDictionary dictionaryWithObjectsAndKeys:branch.path, @"path",
                                branch.mode, @"mode",
                                @"tree", @"type",
                                branch.sha, @"sha",
                                branchUrl, @"url", nil];
  NSDictionary *object = [NSDictionary dictionaryWithObjectsAndKeys:@"test", @"sha",
                          url, @"url",
                          [NSArray arrayWithObjects:leafObject, branchObject, nil], @"tree", nil];
  [tree parser:parser foundObject:object];
  STAssertEquals(2U, [tree.leaves count], @"should have stored the leaves array with the correct number of elements");
  NSMutableArray *types = [NSMutableArray arrayWithObjects:[GHBlob class], [GHTree class], nil];
  for (GHTreeLeaf *newLeaf in tree) {
    [types removeObject:[newLeaf class]];
    if ([newLeaf isKindOfClass:[GHBlob class]]) {
      STAssertTrue([leaf.mode isEqualToString:newLeaf.mode], @"should have stored mode in the blob correctly");
      STAssertTrue([leaf.path isEqualToString:newLeaf.path], @"should have stored path in the blob correctly");
      STAssertTrue([leaf.sha isEqualToString:newLeaf.sha], @"should have stored sha in the blob correctly");
      STAssertEquals(leaf.size, newLeaf.size, @"should have stored size in the blob correctly");
      STAssertTrue([@"blob" isEqualToString:newLeaf.type], @"should have stored type in the blob correctly");
    } else {
      STAssertTrue([branch.mode isEqualToString:newLeaf.mode], @"should have stored mode in the tree correctly");
      STAssertTrue([branch.path isEqualToString:newLeaf.path], @"should have stored path in the tree correctly");
      STAssertTrue([branch.sha isEqualToString:newLeaf.sha], @"should have stored sha in the tree correctly");
      STAssertEquals(branch.size, newLeaf.size, @"should have stored size in the tree correctly");
      STAssertTrue([@"tree" isEqualToString:newLeaf.type], @"should have stored type in the tree correctly");
    }
  }
  STAssertTrue([types count] == 0, @"should have had a blob and a tree");
  STAssertTrue([@"test" isEqualToString:tree.sha], @"should have updated sha");
}

- (void)testParserFoundObject_TreeWithNSNullAttributes_CorrectAttributes {
  STAssertTrue([tree.leaves count] > 0U, @"should have leaves in the tree");
  id parser = [OCMockObject mockForClass:[SBJsonStreamParser class]];
  [tree parser:parser foundObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSNull null], @"sha",
                                   [NSNull null], @"url", [NSNull null], @"tree", nil]];
  STAssertEquals(0U, [tree.leaves count], @"should have erased leaves");
  STAssertNil(tree.sha, @"should have erased sha");
}

@end
